"""Common functions for PyMINEr"""
import pickle
import os
import numpy as np
from scipy.sparse import csc_matrix, lil_matrix
import networkx as nx
import community
import time
import shutil
import h5py
from subprocess import check_call, Popen, PIPE


def cp(file1,file2):
    with open(file1, 'rb') as f_in:
        with open(file2, 'wb') as f_out:
            shutil.copyfileobj(f_in, f_out)
    return()


def rm(rm_file):
    if os.path.isfile(rm_file):
        os.remove(rm_file)
    else:
        print('WARNING:',rm_file,"doesn't exist, couldn't remove it")
    return()


def process_dir(in_dir):
	## process the output dir
	if in_dir[-1]!='/':
		in_dir+='/'
	if not os.path.isdir(in_dir):
		os.makedirs(in_dir)
	return(in_dir)


def run_cmd(in_message, com=True, stdout=None):
    if type(in_message)==str:
        print(in_message)
    elif type(in_message)==list:
        print('\n', " ".join(in_message), '\n')
    else:
        sys.exit("not sure how to parse this for command line run:\n",in_message)
    if stdout:
        with open(stdout, 'w') as out:
            process = Popen(in_message, stdout=PIPE)
            while True:
                line = process.stdout.readline().decode("utf-8")
                out.write(line)
                if line == '' and process.poll() is not None:
                    break
    if com:
        Popen(in_message).communicate()
    else:
        check_call(in_message)


def cmd(in_message, com=True):
    print(in_message)
    time.sleep(.25)
    if com:
        Popen(in_message,shell=True).communicate()
    else:
        Popen(in_message,shell=True)



def read_file(temp_file, lines_o_raw='lines', quiet=False):
    """ basic function library """
    lines = None
    if not quiet:
        print('reading', temp_file)
    file_handle = open(temp_file, 'r')
    if lines_o_raw == 'lines':
        lines = file_handle.readlines()
        for i, line in enumerate(lines):
            lines[i] = line.strip('\n')
    elif lines_o_raw == 'raw':
        lines = file_handle.read()
    file_handle.close()
    return lines


def make_file(contents, path):
    file_handle = open(path, 'w')
    if isinstance(contents, list):
        file_handle.writelines(contents)
    elif isinstance(contents, str):
        file_handle.write(contents)
    file_handle.close()


# def flatten_2D_table(table, delim):
#     # print(type(table))
#     if str(type(table)) == "<class 'numpy.ndarray'>":
#         out = []
#         for i, row in enumerate(table):
#             out.append([])
#             for j, cell in enumerate(row):
#                 try:
#                     str(cell)
#                 except ValueError:
#                     print(cell)
#                 else:
#                     out[i].append(str(cell))
#             out[i] = delim.join(out[i]) + '\n'
#         return out
#     else:
#         for i, row in enumerate(table):
#             for j, cell in enumerate(row):
#                 try:
#                     str(row)
#                 except ValueError:
#                     print(row)
#                 else:
#                     table[i][j] = str(cell)
#             table[i] = delim.join(row) + '\n'
#         return table

def flatten_2D_table(table, delim):
    # print(type(table))
    if str(type(table)) == "<class 'numpy.ndarray'>":
        out = []
        for i in range(0, len(table)):
            out.append([])
            for j in range(0, len(table[i])):
                try:
                    str(table[i][j])
                except:
                    print(table[i][j])
                else:
                    out[i].append(str(table[i][j]))
            out[i] = delim.join(out[i]) + '\n'
        return out
    else:
        for i in range(0, len(table)):
            for j in range(0, len(table[i])):
                try:
                    str(table[i][j])
                except:
                    print(table[i][j])
                else:
                    table[i][j] = str(table[i][j])
            table[i] = delim.join(table[i]) + '\n'
        # print(table[0])
        return table


def strip_split(line, delim='\t'):
    return line.strip('\n').split(delim)


def make_table(lines, delim, range_min=0, num_convert = float):
    for i in range(range_min, len(lines)):
        lines[i] = lines[i].strip()
        lines[i] = lines[i].split(delim)
        for j in range(range_min, len(lines[i])):
            if i!=0 or j!=0:
                try:
                    float(lines[i][j])
                except ValueError:
                    lines[i][j] = lines[i][j].replace('"', '')
                else:
                    lines[i][j] = num_convert(lines[i][j])
    return lines


def get_file_path(in_path):
    in_path = in_path.split('/')
    in_path = in_path[:-1]
    in_path = '/'.join(in_path)
    return in_path + '/'


def read_table(file, sep='\t', num_convert = float):
    return make_table(read_file(file, 'lines'), sep, num_convert = num_convert)


def write_table(table, out_file, sep='\t'):
    make_file(flatten_2D_table(table, sep), out_file)


def import_dict(file_handle):
    file_handle = open(file_handle, 'rb')
    data = pickle.load(file_handle)
    file_handle.close()
    return data


def save_dict(data, path):
    file_handle = open(path, 'wb')
    pickle.dump(data, file_handle)
    file_handle.close()


def check_infile(infile):
    if os.path.isfile(infile):
        return
    else:
        sys.exit(str('could not find '+infile))

def outfile_exists(outfile):
    if os.path.isfile(outfile):
        statinfo = os.stat(outfile)
        if statinfo.st_size!=0:
            return(True)
        else:
            return(False)
    else:
        return(False)




# def get_sample_k_lists(group_numeric_vector, total_groups = None):
#     if total_groups == None:
#         total_groups = max(group_numeric_vector)+1
#     print(total_groups)
#     group_index_lists = [[] for i in range(total_groups)]
#     for i in range(0,len(group_numeric_vector)):
#         temp_group = group_numeric_vector[i]
#         group_index_lists[temp_group].append(i)
#     print("\n\ngroup_index_lists")
#     print(group_index_lists)
#     return(group_index_lists)


def get_sample_k_lists(group_numeric_vector, total_groups = None):
    temp_group_numeric_vector = []
    for num in group_numeric_vector:
        try:
            temp_group_numeric_vector.append(int(num))
        except:
            temp_group_numeric_vector.append(num)
    group_numeric_vector = temp_group_numeric_vector
    if total_groups == None:
        all_vect = list(set(group_numeric_vector))
        numeric_vect = []
        for entry in all_vect:
            #print(type(entry),entry)
            if type(entry)==int:
                numeric_vect.append(entry)
        total_groups = max([len(numeric_vect),max(numeric_vect)+1])
    # print(total_groups)
    group_index_lists = [[] for i in range(total_groups)]
    for i in range(0,len(group_numeric_vector)):
        if type(group_numeric_vector[i]) ==  int or type(group_numeric_vector[i]) == float:

            temp_group = group_numeric_vector[i]
            group_index_lists[temp_group].append(i)
    # print("\n\ngroup_index_lists\n\n")
    # print(group_index_lists)
    return(group_index_lists)


def convert_vect_to_mi_bins(in_vect):
    out_vect = np.array(in_vect, dtype = str)
    ## first get all the unique values
    all_vals = list(set(list(in_vect)))
    ## next get all the numeric convertable values
    ## or return the original vect if there are none
    all_num_vals = []
    all_num_idxs=[]
    for i in range(len(in_vect)):
        try:
            float(in_vect[i])
        except:
            pass
        else:
            temp_val = float(in_vect[i])
            if not np.isnan(temp_val):
                all_num_vals.append(temp_val)
                all_num_idxs.append(i)
    ## if there aren't enough numeric values to warrant conversion, then don't
    if len(all_num_vals) < 10:
        return in_vect
    ## calculate the number of bins
    bin_num = max([2,int(np.sqrt(len(all_num_vals)/5))])
    ## min-max linear normalize the vector
    all_num_vals = np.array(all_num_vals)
    all_num_vals -= np.min(all_num_vals)
    all_num_vals = all_num_vals / np.max(all_num_vals)
    ## calculate the bin interval
    bins = np.arange(bin_num)/bin_num
    digitized_num_vals = np.digitize(all_num_vals, bins)
    print(digitized_num_vals)
    ## now update the original vector to return it
    for i in range(0,len(all_num_idxs)):
        original_idx = all_num_idxs[i]
        new_val = digitized_num_vals[i]
        out_vect[original_idx] = str(new_val)
    return out_vect


def digitize_for_max_info(in_mat):
    ## takes in a full input matrix & digitizes the numeric values of the colums
    in_mat = np.array(in_mat,dtype=str)
    ncol = np.shape(in_mat)[1]
    nrow = np.shape(in_mat)[0]
    for i in range(1,ncol):
        in_mat[1:nrow,i] = convert_vect_to_mi_bins(in_mat[1:nrow,i])
    return in_mat


## this function was adopted from emre's stackoverflow answer found here:
## https://stackoverflow.com/questions/7450957/how-to-implement-rs-p-adjust-in-python
def correct_pvalues_for_multiple_testing(pvalues, correction_type = "Benjamini-Hochberg"):                
    """                                                                                                   
    consistent with R - print correct_pvalues_for_multiple_testing([0.0, 0.01, 0.029, 0.03, 0.031, 0.05, 0.069, 0.07, 0.071, 0.09, 0.1]) 
    """
    from numpy import array, empty
    pvalues = array(pvalues)

    ## convert to linear if needed
    #print(pvalues)
    if len(np.shape(pvalues)) > 1:
        needs_reshaping = True
        original_shape = np.shape(pvalues)
        new_shape = pvalues.size
        #print(new_shape)
        pvalues = pvalues.reshape((new_shape))
        #print(pvalues)
    else:
        needs_reshaping = False

    n = int(pvalues.shape[0])
    new_pvalues = empty(n)
    if correction_type == "Bonferroni":
        new_pvalues = n * pvalues
    elif correction_type == "Bonferroni-Holm":
        values = [ (pvalue, i) for i, pvalue in enumerate(pvalues) ]
        values.sort()
        for rank, vals in enumerate(values):
            pvalue, i = vals
            new_pvalues[i] = (n-rank) * pvalue
    elif correction_type == "Benjamini-Hochberg":
        values = [ (pvalue, i) for i, pvalue in enumerate(pvalues) ]
        values.sort()
        values.reverse()
        new_values = []
        for i, vals in enumerate(values):
            rank = n - i
            pvalue, index = vals
            new_values.append((n/rank) * pvalue)
        for i in range(0, int(n)-1):
            if new_values[i] < new_values[i+1]:
                new_values[i+1] = new_values[i]
        for i, vals in enumerate(values):
            pvalue, index = vals
            new_pvalues[index] = new_values[i]
    if needs_reshaping:
        #print('original_shape\n',pvalues.reshape((original_shape)))
        #print('new_pvalues\n',new_pvalues.reshape((original_shape)))
        new_pvalues = new_pvalues.reshape((original_shape))
    #print("new_pvalues\n",new_pvalues)
    return new_pvalues


def do_louvain_merger(in_mat, sample_k_lists, cutoff = 0):
    ## take in the BH corrected p-values for whether 
    ## the groups are significantly different for this ome
    ## at the level of the feature selected affinity matrix
    # actually cut the edges that need to be pruned (apparently setting the weights to 0 doesn't actually give it zero weight...?)
    cut_counter = 0
    for i in range(in_mat.shape[0]):
        for j in range(in_mat.shape[1]):
            if in_mat[i,j] < cutoff:
                in_mat[i,j] = 0
                cut_counter +=1
    print('cut',cut_counter,'low weight edges')
    ## make the fully connected, weighted matrix
    G = nx.from_numpy_matrix(in_mat)
    for node in G.nodes():
        print(node)
    for edge in G.edges():
        print(edge)
    partition = community.best_partition(G)
    ## this is a dictionary with the original group index as the keys
    ## and the new group index as the 
    print(partition)
    ## make the empty lists for holding all of the 
    ## new sample group lists
    new_groups = []
    new_group_annotations = []
    for i in range(len(set(partition.values()))):
        new_groups.append([])
        new_group_annotations.append([])
    for i in range(in_mat.shape[0]):
        new_groups[partition[i]]+=sample_k_lists[i]
        new_group_annotations[partition[i]].append(i)
    return(new_groups,new_group_annotations, partition)



def removeNonAscii(s):
    ## taken from: https://drumcoder.co.uk/blog/2012/jul/13/removing-non-ascii-chars-string-python/
    return "".join(i for i in s if ord(i)<128)


##################################################################
############### some ray specific functions ######################
def get_num_rows_from_dict_lists(dict_list):
    ## goes through all of the indices & returns the number of dims
    row_dims = 0
    for temp_dict in dict_list:
        for temp_key in list(temp_dict.keys()):
            #print(temp_key)
            if temp_key > row_dims:
                row_dims = temp_key
    return(row_dims+1)


def get_num_cols_from_dict_lists(dict_list):
    ## first get the dimentions
    first_key = list(dict_list[0].keys())[0]
    first = dict_list[0][first_key]
    if type(first) == list:
        col_dims = len(first)
    else:
        col_dims = 1
    return(col_dims)


def ray_dicts_to_array(dict_list, csc = False):
    row_dims = get_num_rows_from_dict_lists(dict_list)
    col_dims = get_num_cols_from_dict_lists(dict_list)
    if csc:
        out_array = np.zeros((row_dims, col_dims))
    else:
        out_array = csc_matrix((row_dims, col_dims))
    print(out_array)
    for temp_dict in dict_list:
        for idx, value in temp_dict.items():
            #print(idx)
            #print(value)
            out_array[idx] = value
    return(out_array)


def get_indices(threads, num_genes):
    indices_list = []
    for t in range(threads):
        indices_list.append([])
    temp_idx = 0
    while temp_idx < num_genes:
        for t in range(threads):
            if temp_idx < num_genes:
                indices_list[t].append(temp_idx)
                temp_idx += 1
    return(indices_list)


def ray_get_indices_from_list(threads, original_idx_list):
    indices_list = []
    for t in range(threads):
        indices_list.append([])
    temp_idx = 0
    while temp_idx < len(original_idx_list):
        for t in range(threads):
            if temp_idx < len(original_idx_list):
                indices_list[t].append(original_idx_list[temp_idx])
                temp_idx += 1
    return(indices_list)
##########################################################

def read_cz_h5(in_file, load_binary=False):
    h5f = h5py.File(in_file , 'r')
    if 'X' not in h5f:
        print("couldn't find the data in the cz h5 file")
    if 'shape' in dir(h5f['X']):
        in_mat= lil_matrix(h5f['X'])
    elif 'keys' in dir(h5f['X']):
        if ('data' in h5f['X']) and ('indices' in h5f['X'])  and ('indptr' in h5f['X']):
            if not load_binary:
                in_mat= csc_matrix((h5f['X']['data'], h5f['X']['indices'], h5f['X']['indptr']))
            else:
                data_vect=np.ones(h5f['X']['data'].shape)
                in_mat= csc_matrix((data_vect, h5f['X']['indices'], h5f['X']['indptr']))
    ## read the genes
    if "shape" in dir(h5f['var']):
        genes = [entry[0] for entry in h5f['var']]
    ## read the cells
    if "shape" in dir(h5f['obs']):
        cells = ['variable'] + [entry[0] for entry in h5f['obs']]
    h5f.close()
    return(in_mat, genes, cells)


def read_tsv(in_file):
    in_table = read_table(in_file)
    cells = in_table[0]
    genes = []
    in_mat_num = np.zeros((len(in_table)-1,len(cells)-1))
    for i in range(1,len(in_table)):
        genes.append(in_table[i,0])
        in_mat_num[i-1,:] = in_table[i,1:]
    return(in_mat, genes, cells)



#!/usr/bin/env python3
from __future__ import division, print_function, absolute_import

import warnings
import math
from collections import namedtuple

# Scipy imports.
from scipy._lib.six import callable, string_types, xrange
from scipy._lib._version import NumpyVersion
from numpy import array, asarray, ma, zeros, var, average
import multiprocessing
from multiprocessing.dummy import Pool as ThreadPool
import scipy.special as special
import scipy.linalg as linalg
import numpy as np
from scipy.stats import rankdata
from copy import deepcopy
import h5py

from scipy.stats import mstats_basic, rankdata

#from . import distributions
#from . import mstats_basic
#from ._distn_infrastructure import _lazywhere
#from ._stats_mstats_common import _find_repeats, linregress, theilslopes
#from ._stats import _kendall_condis

__all__ = []

# __all__ = ['find_repeats', 'gmean', 'hmean', 'mode', 'tmean', 'tvar',
#            'tmin', 'tmax', 'tstd', 'tsem', 'moment', 'variation',
#            'skew', 'kurtosis', 'describe', 'skewtest', 'kurtosistest',
#            'normaltest', 'jarque_bera', 'itemfreq',
#            'scoreatpercentile', 'percentileofscore', 'histogram',
#            'histogram2', 'cumfreq', 'relfreq', 'obrientransform',
#            'signaltonoise', 'sem', 'zmap', 'zscore', 'iqr', 'threshold',
#            'sigmaclip', 'trimboth', 'trim1', 'trim_mean', 'f_oneway',
#            'pearsonr', 'fisher_exact', 'spearmanr', 'pointbiserialr',
#            'kendalltau', 'linregress', 'theilslopes', 'ttest_1samp',
#            'ttest_ind', 'ttest_ind_from_stats', 'ttest_rel', 'kstest',
#            'chisquare', 'power_divergence', 'ks_2samp', 'mannwhitneyu',
#            'tiecorrect', 'ranksums', 'kruskal', 'friedmanchisquare',
#            'chisqprob', 'betai',
#            'f_value_wilks_lambda', 'f_value', 'f_value_multivariate',
#            'ss', 'square_of_sums', 'fastsort', 'rankdata',
#            'combine_pvalues', ]


def _chk_asarray(a, axis):
    if axis is None:
        a = np.ravel(a)
        outaxis = 0
    else:
        a = np.asarray(a)
        outaxis = axis
    if a.ndim == 0:
        a = np.atleast_1d(a)
    return a, outaxis


def _chk2_asarray(a, b, axis):
    if axis is None:
        a = np.ravel(a)
        b = np.ravel(b)
        outaxis = 0
    else:
        a = np.asarray(a)
        b = np.asarray(b)
        outaxis = axis
    if a.ndim == 0:
        a = np.atleast_1d(a)
    if b.ndim == 0:
        b = np.atleast_1d(b)
    return a, b, outaxis


def _contains_nan(a, nan_policy='propagate'):
    policies = ['propagate', 'raise', 'omit']
    if nan_policy not in policies:
        raise ValueError("nan_policy must be one of {%s}" %
                         ', '.join("'%s'" % s for s in policies))
    try:
        # Calling np.sum to avoid creating a huge array into memory
        # e.g. np.isnan(a).any()
        with np.errstate(invalid='ignore'):
            contains_nan = np.isnan(np.sum(a))
    except TypeError:
        # If the check cannot be properly performed we fallback to omiting
        # nan values and raising a warning. This can happen when attempting to
        # sum things that are not numbers (e.g. as in the function `mode`).
        contains_nan = False
        nan_policy = 'omit'
        warnings.warn("The input array could not be properly checked for nan "
                      "values. nan values will be ignored.", RuntimeWarning)
    if contains_nan and nan_policy == 'raise':
        raise ValueError("The input contains nan values")
    return (contains_nan, nan_policy)

def make_dense(b):
    if b is not None:
        if "todense" in dir(b):
            b=b.todense()
    return(b)


def no_p_spear(a, b=None, axis=0, nan_policy='propagate', rm_nans_by_col = False, prop = False):
    a=make_dense(a)
    b=make_dense(b)

    a, axisout = _chk_asarray(a, axis)
    
    contains_nan, nan_policy = _contains_nan(a, nan_policy)

    if contains_nan and rm_nans_by_col:
        ### FINISH THIS
        rm_idxs = np.isna(a)
    
    if contains_nan and nan_policy == 'omit':
        a = ma.masked_invalid(a)
        b = ma.masked_invalid(b)
        return no_p_spear(a, b, axis)
    
    if a.size <= 1:
        return SpearmanrResult(np.nan, np.nan)
    ar = np.apply_along_axis(rankdata, axisout, a)
    
    br = None
    if b is not None:
        b, axisout = _chk_asarray(b, axis)
        
        contains_nan, nan_policy = _contains_nan(b, nan_policy)
        
        if contains_nan and nan_policy == 'omit':
            b = ma.masked_invalid(b)
            return no_p_spear(a, b, axis)
        
        br = np.apply_along_axis(rankdata, axisout, b)
    n = a.shape[axisout]
    if not prop:
        rs = np.corrcoef(ar, br, rowvar=axisout)
    else:
        rs = pp_equivalent_to_cov(ar, br)
    
    olderr = np.seterr(divide='ignore')  # rs can have elements equal to 1
    try:
        # clip the small negative values possibly caused by rounding
        # errors before taking the square root
        t = rs * np.sqrt(((n-2)/((rs+1.0)*(1.0-rs))).clip(0))
    finally:
        np.seterr(**olderr)
    
    if rs.shape == (2, 2):
        return rs[1, 0]
    else:
        return rs


def fast_single_spear(a,b,shuffle = False):
    ## first remove nans
    a_keep = np.logical_not(np.isnan(a))
    b_keep = np.logical_not(np.isnan(b))
    keep = a_keep * b_keep
    a=a[keep]
    b=b[keep]
    ar = rankdata(a,method='dense')
    br = rankdata(b,method='dense')
    if shuffle:
        s_ar = deepcopy(ar)
        np.random.shuffle(s_ar)
        return(np.corrcoef(ar,br)[1,0], np.corrcoef(s_ar,br)[1,0])
    else:
        return(np.corrcoef(ar,br)[1,0])
# def no_p_spear_pairwise(a, b=None, axis=0, nan_policy='propagate', rm_nans_by_col = False, prop = False):
#     ## axis: 1 is row, by row, 0 is column by column
#     if rm_nans_by_col:


#     ar = np.apply_along_axis(rankdata, axisout, a)    
#     br = np.apply_along_axis(rankdata, axisout, b)

#     n = a.shape[axisout]
#     if not prop:
#         rs = np.corrcoef(ar, br, rowvar=axisout)
#     else:
#         rs = pp_equivalent_to_cov(ar, br)
    
#     olderr = np.seterr(divide='ignore')  # rs can have elements equal to 1
#     try:
#         # clip the small negative values possibly caused by rounding
#         # errors before taking the square root
#         t = rs * np.sqrt(((n-2)/((rs+1.0)*(1.0-rs))).clip(0))
#     finally:
#         np.seterr(**olderr)
    
#     if rs.shape == (2, 2):
#         return rs[1, 0]
#     else:
#         return rs


########################################################
########################################################
########################################################
####### proportaionality statistic functions ###########
########################################################
########################################################
########################################################

dummy_array = None
def multi_pp(i):
    global dummy_array
    return(var(dummy_array[i,:] - dummy_array[:,:],axis = 1))

def get_pp(in_mat, verbose = False, do_multi = True):
    ## first get the variance of all genes
    if do_multi:
        threads = multiprocessing.cpu_count()
    else:
        threads = 1
    all_var = var(in_mat, axis = 1)
    if verbose:
        print(all_var)
    ## then get the distance matrix
    ## TODO - make multi-threaded version!
    if do_multi and threads > 1:
        global dummy_array
        dummy_array = in_mat
        indices = list(range(0,in_mat.shape[0]))
        ## set up multi-threaded pool
        pool = ThreadPool(threads)
        all_diff_var = pool.map(multi_pp,indices)
        pool.close()
        pool.join()
        all_diff_var = np.array(all_diff_var)
    else:
        all_diff_var = np.zeros((in_mat.shape[0],in_mat.shape[0]))
        for i in range(0,in_mat.shape[0]):
            all_diff_var[i,:] = var(in_mat[i,:] - in_mat[:,:],axis = 1)
    #all_diff_var = var(in_mat[:,:] - in_mat[:,None,:],axis = 2)
    if verbose:
        print("all_diff_var")
        print(all_diff_var)
    ## then get the pairwise added variance of the original vectors
    all_added_var = all_var + all_var[:,None]
    if verbose:
        print(all_added_var)
    ## then just divide the two matrices by each other
    pp_mat = 1 - (all_diff_var/all_added_var)
    return pp_mat


def get_pp_2(in_mat_a, in_mat_b, verbose = False):
    in_mat_a=make_dense(in_mat_a)
    in_mat_b=make_dense(in_mat_b)

    ## first get the variance of all genes
    all_var_a = var(in_mat_a, axis = 1)
    all_var_b = var(in_mat_b, axis = 1)
    ## then get the distance matrix
    all_diff_var = np.zeros((in_mat_a.shape[0],in_mat_b.shape[0]))
    for i in range(0,in_mat_a.shape[0]):
        all_diff_var[i,:] = var(in_mat_a[i,:] - in_mat_b[:,:],axis = 1)
    #all_diff_var = var(in_mat[:,:] - in_mat[:,None,:],axis = 2)
    if verbose:
        print("all_diff_var")
        print(all_diff_var)
    ## then get the pairwise added variance of the original vectors
    all_added_var = all_var_a[:,None] + all_var_b[None,:]
    if verbose:
        print(all_added_var)
    ## then just divide the two matrices by each other
    pp_mat = 1 - (all_diff_var/all_added_var)
    return pp_mat


def pp_equivalent_to_cov(m, y=None, rowvar=True, bias=False, ddof=None, fweights=None,
        aweights=None):
    """
    Estimate a covariance matrix, given data and weights.
    Covariance indicates the level to which two variables vary together.
    If we examine N-dimensional samples, :math:`X = [x_1, x_2, ... x_N]^T`,
    then the covariance matrix element :math:`C_{ij}` is the covariance of
    :math:`x_i` and :math:`x_j`. The element :math:`C_{ii}` is the variance
    of :math:`x_i`.
    See the notes for an outline of the algorithm.
    Parameters
    ----------
    m : array_like
        A 1-D or 2-D array containing multiple variables and observations.
        Each row of `m` represents a variable, and each column a single
        observation of all those variables. Also see `rowvar` below.
    y : array_like, optional
        An additional set of variables and observations. `y` has the same form
        as that of `m`.
    rowvar : bool, optional
        If `rowvar` is True (default), then each row represents a
        variable, with observations in the columns. Otherwise, the relationship
        is transposed: each column represents a variable, while the rows
        contain observations.
    bias : bool, optional
        Default normalization (False) is by ``(N - 1)``, where ``N`` is the
        number of observations given (unbiased estimate). If `bias` is True,
        then normalization is by ``N``. These values can be overridden by using
        the keyword ``ddof`` in numpy versions >= 1.5.
    ddof : int, optional
        If not ``None`` the default value implied by `bias` is overridden.
        Note that ``ddof=1`` will return the unbiased estimate, even if both
        `fweights` and `aweights` are specified, and ``ddof=0`` will return
        the simple average. See the notes for the details. The default value
        is ``None``.
        .. versionadded:: 1.5
    fweights : array_like, int, optional
        1-D array of integer frequency weights; the number of times each
        observation vector should be repeated.
        .. versionadded:: 1.10
    aweights : array_like, optional
        1-D array of observation vector weights. These relative weights are
        typically large for observations considered "important" and smaller for
        observations considered less "important". If ``ddof=0`` the array of
        weights can be used to assign probabilities to observation vectors.
        .. versionadded:: 1.10
    Returns
    -------
    out : ndarray
        The covariance matrix of the variables.
    See Also
    --------
    corrcoef : Normalized covariance matrix
    Notes
    -----
    Assume that the observations are in the columns of the observation
    array `m` and let ``f = fweights`` and ``a = aweights`` for brevity. The
    steps to compute the weighted covariance are as follows::
        >>> m = np.arange(10, dtype=np.float64)
        >>> f = np.arange(10) * 2
        >>> a = np.arange(10) ** 2.
        >>> ddof = 9 # N - 1
        >>> w = f * a
        >>> v1 = np.sum(w)
        >>> v2 = np.sum(w * a)
        >>> m -= np.sum(m * w, axis=None, keepdims=True) / v1
        >>> cov = np.dot(m * w, m.T) * v1 / (v1**2 - ddof * v2)
    Note that when ``a == 1``, the normalization factor
    ``v1 / (v1**2 - ddof * v2)`` goes over to ``1 / (np.sum(f) - ddof)``
    as it should.
    Examples
    --------
    Consider two variables, :math:`x_0` and :math:`x_1`, which
    correlate perfectly, but in opposite directions:
    >>> x = np.array([[0, 2], [1, 1], [2, 0]]).T
    >>> x
    array([[0, 1, 2],
           [2, 1, 0]])
    Note how :math:`x_0` increases while :math:`x_1` decreases. The covariance
    matrix shows this clearly:
    >>> np.cov(x)
    array([[ 1., -1.],
           [-1.,  1.]])
    Note that element :math:`C_{0,1}`, which shows the correlation between
    :math:`x_0` and :math:`x_1`, is negative.
    Further, note how `x` and `y` are combined:
    >>> x = [-2.1, -1,  4.3]
    >>> y = [3,  1.1,  0.12]
    >>> X = np.stack((x, y), axis=0)
    >>> np.cov(X)
    array([[11.71      , -4.286     ], # may vary
           [-4.286     ,  2.144133]])
    >>> np.cov(x, y)
    array([[11.71      , -4.286     ], # may vary
           [-4.286     ,  2.144133]])
    >>> np.cov(x)
    array(11.71)
    """
    # Check inputs
    if ddof is not None and ddof != int(ddof):
        raise ValueError(
            "ddof must be integer")

    # Handles complex arrays too
    m = np.asarray(m)
    if m.ndim > 2:
        raise ValueError("m has more than 2 dimensions")

    if y is None:
        dtype = np.result_type(m, np.float64)
    else:
        y = np.asarray(y)
        if y.ndim > 2:
            raise ValueError("y has more than 2 dimensions")
        dtype = np.result_type(m, y, np.float64)

    X = array(m, ndmin=2, dtype=dtype)
    if not rowvar and X.shape[0] != 1:
        X = X.T
    if X.shape[0] == 0:
        return np.array([]).reshape(0, 0)
    if y is not None:
        y = array(y, copy=False, ndmin=2, dtype=dtype)
        if not rowvar and y.shape[0] != 1:
            y = y.T
        X = np.concatenate((X, y), axis=0)

    if ddof is None:
        if bias == 0:
            ddof = 1
        else:
            ddof = 0

    # Get the product of frequencies and weights
    w = None
    if fweights is not None:
        fweights = np.asarray(fweights, dtype=float)
        if not np.all(fweights == np.around(fweights)):
            raise TypeError(
                "fweights must be integer")
        if fweights.ndim > 1:
            raise RuntimeError(
                "cannot handle multidimensional fweights")
        if fweights.shape[0] != X.shape[1]:
            raise RuntimeError(
                "incompatible numbers of samples and fweights")
        if any(fweights < 0):
            raise ValueError(
                "fweights cannot be negative")
        w = fweights
    if aweights is not None:
        aweights = np.asarray(aweights, dtype=float)
        if aweights.ndim > 1:
            raise RuntimeError(
                "cannot handle multidimensional aweights")
        if aweights.shape[0] != X.shape[1]:
            raise RuntimeError(
                "incompatible numbers of samples and aweights")
        if any(aweights < 0):
            raise ValueError(
                "aweights cannot be negative")
        if w is None:
            w = aweights
        else:
            w *= aweights

    avg, w_sum = average(X, axis=1, weights=w, returned=True)
    w_sum = w_sum[0]

    # Determine the normalization
    if w is None:
        fact = X.shape[1] - ddof
    elif ddof == 0:
        fact = w_sum
    elif aweights is None:
        fact = w_sum - ddof
    else:
        fact = w_sum - ddof*sum(w*aweights)/w_sum

    if fact <= 0:
        warnings.warn("Degrees of freedom <= 0 for slice",
                      RuntimeWarning, stacklevel=3)
        fact = 0.0

    X -= avg[:, None]
    if w is None:
        X_T = X.T
    else:
        X_T = (X*w).T
    c = get_pp(X)
    c *= np.true_divide(1, fact)
    return c.squeeze()


########################################################
########################################################
########################################################


def get_Z_cutoff(vect, z=4.5,positive = False):
    ## this is for a half distribution, so mult by -1 to make it 'normal'
    vect_original = np.array(deepcopy(vect))
    keep = np.logical_not(np.isnan(vect_original))
    vect_original = vect_original[keep]
    vect = deepcopy(vect_original)
    vect = list(vect)
    vect += list(np.array(vect)*-1)
    vect = np.array(vect)
    ## get the mean, & SD
    v_mean = np.mean(vect)
    v_std = np.std(vect)
    cutoff = v_std * z
    if not positive:
        cutoff *= -1
        cutoff = max([-0.9,cutoff])
    else:
        cutoff = min([0.9,cutoff])
    print('cutoff:',cutoff)
    if not positive:
        num_sig = np.sum(np.array(vect_original<cutoff,dtype = bool))
        total_num = vect_original.shape[0]
        print(num_sig, "/", total_num, 'significant')
        FPR = max([1,total_num])/num_sig
        return(cutoff, FPR)
    else:
        return(cutoff)



#!/usr/bin/env python3
import pickle, scipy, sys,  fileinput, gc, os
from networkx import *
import numpy as np
import matplotlib.pyplot as plt
from copy import deepcopy
try:
    import community
except:
    community_installed = False
    print('https://github.com/taynaud/python-louvain is a requirement for finding communites')
else:
    import community
    community_installed = True
gc.enable()
######################################################################################
###############
## basic function library
def read_file(tempFile,linesOraw,quiet=False):
    if not quiet:
        print('reading',tempFile)
    f=open(tempFile,'r')
    if linesOraw=='lines':
        lines=f.readlines()
        for i in range(0,len(lines)):
            lines[i]=lines[i].strip('\n')
    elif linesOraw=='raw':
        lines=f.read()
    f.close()
    return(lines)

def make_file(contents,path):
    f=open(path,'w')
    if isinstance(contents,list):
        f.writelines(contents)
    elif isinstance(contents,str):
        f.write(contents)
    f.close()

    
def flatten_2D_table(table,delim):
    #print(type(table))
    if str(type(table))=="<class 'numpy.ndarray'>":
        out=[]
        for i in range(0,len(table)):
            out.append([])
            for j in range(0,len(table[i])):
                try:
                    str(table[i][j])
                except:
                    print(table[i][j])
                else:
                    out[i].append(str(table[i][j]))
            out[i]=delim.join(out[i])+'\n'
        return(out)
    else:
        for i in range(0,len(table)):
            for j in range(0,len(table[i])):
                try:
                    str(table[i][j])
                except:
                    print(table[i][j])
                else:
                    table[i][j]=str(table[i][j])
            table[i]=delim.join(table[i])+'\n'
    #print(table[0])
        return(table)

def make_table(lines,delim):
    for i in range(0,len(lines)):
        lines[i]=lines[i].strip()
        lines[i]=lines[i].split(delim)
        for j in range(0,len(lines[i])):
            try:
                float(lines[i][j])
            except:
                lines[i][j]=lines[i][j].replace('"','')
            else:
                lines[i][j]=float(lines[i][j])
    return(lines)


def read_table(file, sep='\t'):
    return(make_table(read_file(file,'lines'),sep))
    
def write_table(table, out_file, sep = '\t'):
    make_file(flatten_2D_table(table,sep), out_file)


def import_dict(f):
    f=open(f,'rb')
    d=pickle.load(f)
    f.close()
    return(d)

def save_dict(d,path):
    f=open(path,'wb')
    pickle.dump(d,f)
    f.close()

def cmd(in_message, com=True):
    print(in_message)
    time.sleep(.25)
    if com:
        Popen(in_message,shell=True).communicate()
    else:
        Popen(in_message,shell=True)

def strip_split(line, delim = '\t'):
    return(line.strip('\n').split(delim))
############################################################################################
##############################################################
import argparse
parser = argparse.ArgumentParser()

def valid_layout(in_arg):
	if in_arg in ['spring','spectral','circle','random','shell']:
		return(in_arg)
	else:
		msg = '\n'+in_arg+" is not a valid layout, specify one of the following:\n'spring','spectral','circle','random','shell'"
		raise argparse.ArgumentTypeError(msg)

parser.add_argument("-graph",'-g',
    help="if you already have a graph object that was saved by this program, we can just use that instead of recalculating everything",
    type = str)
parser.add_argument("-id_list",'-id','-ID','-IDs','-ids', 
    help="these are sorted IDs matching the order of the original input dataset")
parser.add_argument("-adj_list","-adj", 
    help="this file contains the adjacency list which will make the graph")
parser.add_argument("-rand_seed", 
    help="the random seed used in generating the layout",
    type = int,
    default = 12345678)
parser.add_argument("-no_default_plot", "-no_default",
    help="Don't do the default plot with red nodes",
    action = 'store_true',
    default = False)
parser.add_argument("-node_color","-nc", 
    help="Name of the node attribute(s) (separated by commas) to be used for coloring the nodes. As currently implemented, this only works for continuous variables")
parser.add_argument("-color_scheme",'-cs', 
    help="matplotlib color scheme for the node colors if used",
    default = "bwr")
parser.add_argument("-layout", 
    help="the layout algorithm used",
    type = valid_layout,
    default = "spring")
parser.add_argument("-redo_layout", 
    help="If you already have a graph, but want to re-do the layout",
    action = 'store_true',
    default = False)
parser.add_argument("-layout_iters", 
    help="number of iterations to be used in generating the graph layout",
    type = int,
    default = 35)
parser.add_argument("-k", 
    help="the desired distance between each node",
    type = float,
    default = 0.25)
parser.add_argument("-directed", 
    help="whether or not the graph is directed",
    action = 'store_true',
    default = False)
parser.add_argument("-output", '-out_dir',
    help="the directory to place the graph files in",
    type = str)
parser.add_argument("-node_attribute_lists",'-node_attributes','-node_atts','-node_attrs', 
    help="1 or more files (comma separated list) containing file(s) for logging node attributes",
    type = str)
parser.add_argument("-edge_attribute_lists",'-edge_atts', 
    help="this is the leading string for generating the saved graph, as well as the images",
    type = str)
parser.add_argument("-show_intermediates", 
    help="NOT YET SUPPORTED - whether or not to show the intermediate graphs while the layout algorithm is iterating",
    type = bool)
parser.add_argument("-verbose", '-v',
    help="whether or not to show some verbose trouble shooting lines",
    action = 'store_true')
parser.add_argument("-min_comp_size", 
    help="the minimum size of a component for plotting. Units are percent of dataset in largest component (default = 10.0, must be 0-100)",
    type = float,
    default = 10.)
parser.add_argument("-dpi", 
    help="dots per inch for plotting networks",
    type = int,
    default = 360)
parser.add_argument("-plot_all", 
    help="if you want to plot all of the attributes",
    default = False,
    action = 'store_true')
parser.add_argument("-dont_replot", 
    help="if you want to re-plot the attributes, but don't want to, including those which were already plotted before in the output directory",
    default = True,
    action = 'store_false',
    dest = 'replot')
parser.add_argument("-dont_save", 
    help="if you don't want to save the graphpkl",
    default = True,
    action = 'store_false',
    dest = 'save')
parser.add_argument("-float_color_range",'-range', '-color_range', '-col_range', 
    help="a comma delimited range of two floats which give the low and high end for showing expression levels. This can be used to enhance contrast of an image.",
    type = str)
parser.add_argument("-subset_nodes", 
    help="to plot a specific subset of nodes in the graph",
    type = str)
args = parser.parse_args()

##############################################################
if args.output[-1]!='/':
    args.output+='/'

np.random.seed(args.rand_seed)
import random
random.seed(args.rand_seed)


if args.float_color_range != None:
    if ',' not in args.float_color_range:
        sys.exit('float_color_range must be comma delimited')
    express_range = args.float_color_range.split(',')
    express_range[0] = float(express_range[0])
    express_range[1] = float(express_range[1])
    min_express = min(express_range)
    max_express = max(express_range)
else:
    min_express = None
    max_express = None
#######################################
def get_id_list_if_none():
    ## if we don't have any id list given as an argument, we'll generate the id_list from 
    ## the adj list
    global args
    id_list = []
    first = True
    for line in fileinput.input(args.adj_list):
        if first:
            first = False
        else:
            temp_line = strip_split(line)
            id_list+=temp_line[:2]
    return(sorted(list(set(id_list))))

#######################################
if args.adj_list == None and args.graph == None:
	G = lollipop_graph(4,6)
	num_edge = 6
#######################################
else:
    if args.graph != None:
        ## read in the input graph
        in_graph = import_dict(args.graph)
        G = in_graph['graph']
        ID_list = G.nodes()
        if 'positions' in in_graph:
            pos = in_graph['positions']
            num_node = number_of_nodes(G)
            num_edge = number_of_edges(G)
    else:
		## read in the IDs for setting up the layout of the adj matrix
        if args.id_list == None:
            ID_list = get_id_list_if_none()
        else:
            ID_list = read_file(args.id_list, "lines")
        num_node = len(ID_list)
        if args.directed:
        	G = DiGraph()
        else:
        	G = Graph()
        ID_hash = {}

        ## add the nodes to the graph
        print('\nadding the nodes to the graph\n')
        for i in range(0,len(ID_list)):
        	ID_hash[ID_list[i]] = i
        	G.add_node(ID_list[i])

        ## go through the adj_list and add the edges
        print('\nadding the edges to the graph\n')
        num_edge = 0
        first = True
        for line in fileinput.input(args.adj_list):
        	if first:
        		first = False
        	else:
        		num_edge+=1
        		temp_line = strip_split(line)
        		G.add_edge(temp_line[0],temp_line[1])

args.min_comp_size = len(ID_list)*args.min_comp_size/100
##################################################################
empty_node_dict = {}
all_nodes = G.nodes()
for n in all_nodes:
    empty_node_dict[n]=0

def get_empty_node_dict(G):
    empty_node_dict = {}
    all_nodes = G.nodes()
    for n in all_nodes:
        empty_node_dict[n]=0
    return(empty_node_dict)
##################################################################
## get the comminities based on the louvain method described in:
## Fast unfolding of communities in large networks,
## Vincent D Blondel, Jean-Loup Guillaume, Renaud Lambiotte, Renaud Lefebvre, Journal of Statistical Mechanics: Theory and Experiment 2008(10), P10008 
############################################################
def normalize_PR_dict(pr_dict):
    nodes = list(pr_dict.keys())
    original_pr = []
    for node in nodes:
        original_pr.append(pr_dict[node])
    original_pr_array = np.array(original_pr)
    original_pr_array = original_pr_array - np.min(original_pr_array)
    original_pr_array = original_pr_array / np.max(original_pr_array)
    norm_pr_dict = {}
    for i in range(0,len(nodes)):
        node = nodes[i]
        norm_pr_dict[node] = original_pr_array[i]
    return(norm_pr_dict)


def get_LPR(G, community_dict):
    ## takes in a graph, and it's partitions.
    ## Returns the dictionary of local page ranks
    ## calculate within community page-rank
    print('finding Local PageRank (LPR)')
    LPR_dict = get_empty_node_dict(G)
    all_nodes = list(LPR_dict.keys())
    ## get the list of all unique communities
    all_com_list = set(community_dict.values())
    #print(list(community_dict.keys()))
    all_com_dict = {com:[] for com in all_com_list}
    for node in all_nodes:
        temp_com = community_dict[node]
        temp_node_list = all_com_dict[temp_com]
        temp_node_list.append(node)
        all_com_dict[temp_com] = temp_node_list
    for com in all_com_list:
        temp_nodes = all_com_dict[com]
        if len(temp_nodes)>=5:
            temp_subgraph = G.subgraph(temp_nodes)
            temp_pr_dict = pagerank(temp_subgraph)
            ## now normalize it within this community
            temp_pr_dict = normalize_PR_dict(temp_pr_dict)
            for node in temp_pr_dict.keys():
                LPR_dict[node] = temp_pr_dict[node]
                #print(node,temp_pr_dict[node])
    return(LPR_dict)


############################################################

if community_installed:
    print('\nfinding the communities\n')
    if nx.get_node_attributes(G,'community') == {}:
        try:
            partition = community.best_partition(G)
        except:
            community_installed = False
        else:
            pr = pagerank(G)#,nstart=0.15)
            LPR = get_LPR(G, partition)
            partition_table = [['ID','community', 'Local_PageRank', 'PageRank']]
            for i in ID_list:
                partition[i]='community_'+str(partition[i])
                partition_table.append([i,partition[i], LPR[i], pr[i]])

            try:
                set_node_attributes(G,pr,name = "PageRank")
            except:
                set_node_attributes(G,'PageRank',pr)
            try:
                set_node_attributes(G,partition,name = "community")
            except:
                set_node_attributes(G,'community',partition)
            try:
                set_node_attributes(G,LPR,name = "LPR")
            except:
                set_node_attributes(G,'LPR',LPR)

        write_table(partition_table,args.output+'communities.txt')
    else:
        print('already done!')


##################################################################
print('\nfinding the connected components\n')
comps = connected_components(G)

## figure out which nodes to plot based on the connected component size
big_comp_nodes = []
small_comp_nodes = []
for i in comps:
	if len(i)>=args.min_comp_size:
		big_comp_nodes += i
	else:
		small_comp_nodes += i
print(len(big_comp_nodes),'nodes were found in the components larger than',args.min_comp_size)

large_comp_G = G.subgraph(big_comp_nodes)
#large_comp_G.remove_nodes_from(small_comp_nodes)



##################################################################
####### calculate the positions based on the given layout ########

if 'pos' not in vars().keys() or args.redo_layout:#args.graph == None or args.graph_positons == None:
	print('\ncalculating the node positions in the graph\n(this could take a while)\n')
	layout_function_dict = {'spring':spring_layout,'spectral':spectral_layout,
	'circle':circular_layout,'random':random_layout,'shell':shell_layout}

	layout_alg = layout_function_dict[args.layout]
	if args.layout == 'spring':
		pos=layout_alg(large_comp_G,iterations=args.layout_iters,k=args.k, seed=args.rand_seed)
	else:
		pos=layout_alg(large_comp_G, seed=args.rand_seed)
else:
	pass
gc.collect()
##################################################################
## import node attributes


def process_node(temp_node):
    try:
        float(temp_node)
    except:
        return(temp_node)
    else:
        return(str(float(temp_node)))



if args.node_attribute_lists != None:
    list_of_na_files = args.node_attribute_lists.split(',')
    for f in list_of_na_files:
        temp_attr_table = read_table(f)
        for i in range(1,len(temp_attr_table[0])):
            temp_attr_name = temp_attr_table[0][i]
            temp_attr_name = temp_attr_name.replace('/','|')
            print('\nsetting the node attribute:',temp_attr_name,'\n')
            temp_attr_dict = empty_node_dict.copy()
            print(list(temp_attr_dict.keys())[:5])
            for n in range(1,len(temp_attr_table)):
                temp_node = process_node(temp_attr_table[n][0])
                if temp_node in temp_attr_dict:
                    temp_attr_dict[temp_node] = temp_attr_table[n][i]
                else:
                    ## this means that we couldn't find the node from the
                    ## attr list in the graph
                    if args.verbose:
                        print(temp_node,"not in temp_attr_dict")
                    pass
            try:
                print("temp_attr_name",temp_attr_name)
                set_node_attributes(G,temp_attr_dict,name = temp_attr_name)
            except:
                set_node_attributes(G,temp_attr_name,temp_attr_dict)

##################################################################
## import edge attributes



##################################################################

## node and edge sizes
node_size = 2.5

edge_width = 2.5/np.sqrt(number_of_edges(G))

gc.collect()
##################################################################
## plot the graph
## select some random edges to draw so that it can do it in a reasonable amnt of time
try:
    all_edges=list(large_comp_G.edges)
except:
    all_edges=list(large_comp_G.edges())
#print(all_edges)
num_edge = int(1e6)
if len(all_edges)>num_edge:
    new_order= all_edges[:]
    np.random.shuffle(all_edges)
    subset_edges = new_order[:num_edge]
else:
    subset_edges = all_edges

if args.no_default_plot == False:

    print('\nplotting the graph\n')
    out_graph_plot = draw(large_comp_G,pos, node_shape = 'o', linewidths = 0, 
    	node_size = node_size, width = edge_width, font_size=0)


    if args.output != None:
        ## save the image
        if args.output[-1]!='/':
            args.output+='/'
        plt.savefig(args.output+'full_graph.png',dpi=args.dpi,bbox_inches='tight')
        plt.clf()
    else:
        plt.show()
gc.collect()
##################################################################
## either save or display the graph(s)
##############
def get_color_from_var(in_attr_dict):
    if is_categorical(in_attr_dict):
        return(get_vals_from_dict(get_color_from_cat(in_attr_dict)))
    else:
        return(get_vals_from_dict(get_color_from_float(in_attr_dict)))

def get_vals_from_dict(val_map):
    global large_comp_G
    vals = [val_map[node] for node in large_comp_G.nodes()]
    return(vals)

def get_color_from_cat(in_attr_dict):
    all_vals = []
    for n in sorted(list(in_attr_dict.keys())):
        all_vals.append(in_attr_dict[n])
    unique_entries = sorted(list(set(all_vals)))
    num_vars = len(unique_entries)
    cat_to_float_dict = {}
    for entry in enumerate(unique_entries):
        cat_to_float_dict[entry[1]]=entry[0]/(num_vars)

    ## update the in_attr_dict to be float rather than cat
    for n in in_attr_dict.keys():
        in_attr_dict[n]=cat_to_float_dict[in_attr_dict[n]]

    return(get_color_from_float(in_attr_dict))

def is_categorical(in_attr_dict):
    global large_comp_G
    ## first we need to figure out if it's a float or a categorical variable
    ## the first sign that we have a categorical var is if there are strings
    ## in the attribute dict
    is_categorical_bool = False
    all_vals = []
    for n in in_attr_dict.keys():
        all_vals.append(in_attr_dict[n])
        if type(in_attr_dict[n]) == str:
            is_categorical_bool = True
    
    ## no we can check how variant the vector is
    ## if there are not that many unique values compared to the number of nodes
    ## then it's probably categorical
    ## by default this is set to 10%
    #num_unique = len(list(set(all_vals)))
    #if number_of_nodes(G)/10 > num_unique:
    #    is_categorical_bool = True

    return(is_categorical_bool)


def get_color_from_float(in_attr_dict):
    global min_express, max_express

    if min_express != None:
        print('setting mins to:',min_express)
        print('setting maxs to:',max_express)
        for n in in_attr_dict.keys():
            if in_attr_dict[n] > max_express:
                #print('setting',in_attr_dict[n],'to',max_express)
                in_attr_dict[n] = max_express
            elif in_attr_dict[n] < min_express:
                #print('setting',in_attr_dict[n],'to',min_express)
                in_attr_dict[n] = min_express
            else:
                pass

    #### do linear normalization between 0 and 1
    ## get the min
    
    temp_min = None
    for n in in_attr_dict.keys():
        #print(n)
        #print('temp_min',temp_min)
        if temp_min == None:
            temp_min = in_attr_dict[n]
        else:
            if temp_min > in_attr_dict[n]:
                temp_min = in_attr_dict[n]
                #print('temp_min',temp_min)
            else:
                pass
    ## subtract min
    #print('temp_min',temp_min)
    for n in in_attr_dict.keys():
        in_attr_dict[n] = in_attr_dict[n]-temp_min

    ## get the max
    temp_max = None
    for n in in_attr_dict.keys():
        #print(n)
        #print('temp_max',temp_max)
        if temp_max == None:
            temp_max = in_attr_dict[n]
            #print('temp_max',temp_max)
        else:
            if temp_max < in_attr_dict[n]:
                temp_max = in_attr_dict[n]
                #print('temp_max',temp_max)
            else:
                pass
    # divide max
    #print('temp_max',temp_max)
    for n in in_attr_dict.keys():
        if temp_max == 0:
            in_attr_dict[n] = 0
        else:
            in_attr_dict[n] = in_attr_dict[n]/temp_max

    return(in_attr_dict)

#####################
## get the names of all attributes
def get_attrs():
    global large_comp_G
    for i in iter(large_comp_G.node.items()):
        first_node = i[1]
        break
    return(list(first_node.keys()))
#####################


if args.output != None:
    if args.node_color != None or args.plot_all:
        ## if it's a big graph, we'll only plot a subset of the edges
        ## Not implemented yet
        # all_edges = list(large_comp_G.edges())
        # max_edges = int(min([1e5,len(all_edges)]))
        # sample_edges = random.sample(all_edges,max_edges)
        ## plot the other node colors
        plt.clf()
        if args.node_color != None:
            node_color_list = args.node_color.split(',')
        if args.plot_all:
            node_color_list = get_attrs()
        for nc in node_color_list:
            if args.replot or not os.path.isfile(args.output+nc+'.png'):
                plt.clf()
                print('\nplotting the',nc,'graph')
                #val_map = get_color_from_float(get_node_attributes(G,nc))
                #values = [val_map[node] for node in G.nodes()]
                temp_nodes = list(large_comp_G.nodes())

                if is_categorical(get_node_attributes(large_comp_G,nc)):
                    ## check to see if we've already logged the values for plotting
                    if get_node_attributes(large_comp_G,nc+"_values") == {}:
                        ## if we haven't, then do it
                        values = get_color_from_var(get_node_attributes(large_comp_G,nc))
                        ## now log them
                        new_val_dict = {}
                        for i in range(len(temp_nodes)):
                            new_val_dict[temp_nodes[i]] = values[i]
                        set_node_attributes(large_comp_G, new_val_dict, name = nc+'_values')
                    else:
                        ## if we have then just use those again
                        temp_values_dict = get_node_attributes(large_comp_G,nc+"_values")

                        print('retreiving previously used categorical colors')
                        values = []
                        for i in range(len(temp_nodes)):
                            values.append(temp_values_dict[temp_nodes[i]])
                else:
                    ## if it's not categorical, we can just get the numeric attributes directly
                    values = get_color_from_var(get_node_attributes(large_comp_G,nc))

                #print('\t',values)
                if is_categorical(get_node_attributes(large_comp_G,nc)):
                    print('\tis categorical')
                    temp_colors = 'gist_rainbow'#'pipy_spectral'#'Set1'
                else:
                    print("\tisn't categorical")
                    temp_colors = args.color_scheme
                ## convert the values to colors
                temp_cmap = plt.get_cmap(temp_colors)
                temp_value_colors = []
                for v in values:
                    temp_value_colors.append(temp_cmap(v))
                values = temp_value_colors
                if args.subset_nodes != None:
                    temp_subset = read_file(args.subset_nodes, 'lines')
                    subset_hash = {key:value for value, key in enumerate(temp_subset)}
                    final_subset = []
                    final_values = []
                    for i in range(0,len(temp_nodes)):
                        if temp_nodes[i] in subset_hash:
                            final_subset.append(temp_nodes[i])
                            final_values.append(values[i])
                    temp_nodes = final_subset
                    values = final_values
                draw(large_comp_G,pos, cmap=plt.get_cmap(temp_colors),
                    node_color = values, node_shape = 'o', linewidths = 0, 
                    node_size = node_size, width = edge_width, font_size=0,
                    nodelist=temp_nodes,
                    edgelist=subset_edges)
                #draw_networkx_nodes(G,pos,nodelist=list(G.nodes()),
                #    cmap=plt.get_cmap(temp_colors),
                #    node_color = values,
                #    node_size = node_size,
                #    linewidths = 0)
                ## save the image
                plt.savefig(args.output+nc+'.png',
                    dpi=args.dpi,
                    bbox_inches='tight')
                gc.collect()
            else:
                if os.path.isfile(args.output+nc+'.png'):
                    print('\n',args.output+nc+'.png', 'already exists\nskipping')
                gc.collect()
        gc.collect()

#################################################
## save the graphs
if args.output != None and args.save:
    print('\nsaving the graph\n')
    ## save the graph
    save_dict({'graph':large_comp_G,'positions':pos},args.output+'large_comps.graphpkl')
    
    save_dict({'graph':G},args.output+'full_graph_full.graphpkl')


#!/usr/bin/env python3

##import dependency libraries
import sys,time,glob,os,pickle,fileinput,argparse
from subprocess import Popen
from operator import itemgetter
import gc, fileinput, random
import numpy as np
try:
    from pyminer.common_functions import *
except:
    from common_functions import *
## check that we have gprofiler installed
try:
    from gprofiler import GProfiler
except:
    sys.exit('please install gprofiler; try: pip3 install gprofiler-official==0.3.5')
else:
    from gprofiler import GProfiler

#import pandas as pd
##############################################################
## basic function library
# def read_file(tempFile,linesOraw='lines',quiet=False):
#     if not quiet:
#         print('reading',tempFile)
#     f=open(tempFile,'r')
#     if linesOraw=='lines':
#         lines=f.readlines()
#         for i in range(0,len(lines)):
#             lines[i]=lines[i].strip('\n')
#     elif linesOraw=='raw':
#         lines=f.read()
#     f.close()
#     return(lines)

# def make_file(contents,path):
#     f=open(path,'w')
#     if isinstance(contents,list):
#         f.writelines(contents)
#     elif isinstance(contents,str):
#         f.write(contents)
#     f.close()

    
# def flatten_2D_table(table,delim):
#     #print(type(table))
#     if str(type(table))=="<class 'numpy.ndarray'>":
#         out=[]
#         for i in range(0,len(table)):
#             out.append([])
#             for j in range(0,len(table[i])):
#                 try:
#                     str(table[i][j])
#                 except:
#                     print(table[i][j])
#                 else:
#                     out[i].append(str(table[i][j]))
#             out[i]=delim.join(out[i])+'\n'
#         return(out)
#     else:
#         for i in range(0,len(table)):
#             for j in range(0,len(table[i])):
#                 try:
#                     str(table[i][j])
#                 except:
#                     print(table[i][j])
#                 else:
#                     table[i][j]=str(table[i][j])
#             table[i]=delim.join(table[i])+'\n'
#     #print(table[0])
#         return(table)

# def strip_split(line, delim = '\t'):
#     return(line.strip('\n').split(delim))

# def make_table(lines,delim):
#     for i in range(0,len(lines)):
#         lines[i]=lines[i].strip()
#         lines[i]=lines[i].split(delim)
#         for j in range(0,len(lines[i])):
#             try:
#                 float(lines[i][j])
#             except:
#                 lines[i][j]=lines[i][j].replace('"','')
#             else:
#                 lines[i][j]=float(lines[i][j])
#     return(lines)


# def get_file_path(in_path):
#     in_path = in_path.split('/')
#     in_path = in_path[:-1]
#     in_path = '/'.join(in_path)
#     return(in_path+'/')


# def read_table(file, sep='\t'):
#     return(make_table(read_file(file,'lines'),sep))
    
# def write_table(table, out_file, sep = '\t'):
#     make_file(flatten_2D_table(table,sep), out_file)
    

# def import_dict(f):
#     f=open(f,'rb')
#     d=pickle.load(f)
#     f.close()
#     return(d)

# def save_dict(d,path):
#     f=open(path,'wb')
#     pickle.dump(d,f)
#     f.close()

# all_calls = []
# def cmd(in_message, com=True):
#     global all_calls
#     if type(in_message)==list:
#         in_message = ' '.join(in_message)
#     print(in_message)
#     all_calls.append(in_message)
#     time.sleep(.25)
#     if com:
#         Popen(in_message,shell=True).communicate()
#     else:
#         Popen(in_message,shell=True)

# def check_infile(infile):
#     if os.path.isfile(infile):
#         return
#     else:
#         sys.exit(str('could not find '+infile))

# def outfile_exists(outfile):
#     if os.path.isfile(outfile):
#         statinfo = os.stat(outfile)
#         if statinfo.st_size!=0:
#             return(True)
#         else:
#             return(False)
#     else:
#         return(False)


####################################################################################


###########################
def convert_to_human(in_genes, species):
    ## setup the api
    gp = GProfiler('PyMINEr_'+str(random.randint(0,int(1e6))), want_header = True)

    temp_lookup={}
    gene_symbol_dict = {}
    gene_def_dict = {}
    for g in in_genes:
        temp_lookup[g] = []
        gene_symbol_dict[g] = []
        

    # import inspect
    # code = inspect.getsource(gp.gorth)
    # print(code)
    results = gp.gorth(in_genes, 
        source_organism = species,
        target_organism='hsapiens')




    for i in range(1,len(results)):
        temp_id = results[i][1]
        temp_hu_symbol = results[i][5]
        temp_hu_def = results[i][6]
        if temp_id not in gene_symbol_dict:
            print(temp_id,"to:")
            temp_id = in_genes[int(results[i][0])-1]
            print("\t",temp_id)
        if temp_id not in gene_symbol_dict:
            pass#print('weired mapping event with:',temp_id)
        elif temp_hu_symbol != None:
            temp_symbol_list = gene_symbol_dict[temp_id]
            temp_symbol_list.append(temp_hu_symbol)
            gene_symbol_dict[temp_id] = temp_symbol_list
            gene_def_dict[temp_id] = temp_hu_def



    return(gene_symbol_dict,gene_def_dict,results)


def convert_to_ensg(in_genes, species = 'hsapiens'):

    ## setup the api
    gp = GProfiler('PyMINEr_'+str(random.randint(0,int(1e6))), want_header = True)

    temp_lookup={}
    gene_symbol_dict = {}
    gene_def_dict = {}
    for g in in_genes:
        temp_lookup[g] = []
        gene_symbol_dict[g] = []
        gene_def_dict[g] = []


    results = gp.gconvert(in_genes, 
        organism = species, target='ENSG',numeric_ns="ENTREZGENE_ACC")


    for i in range(1,len(results)):
        temp_id = results[i][1]
        temp_id = temp_id.replace("ENTREZGENE_ACC:","")
        temp_hu_symbol = results[i][4]
        temp_def = results[i][5]
        if temp_id not in gene_symbol_dict:
            print(temp_id,"to:")
            temp_id = in_genes[int(results[i][0])-1]
            print("\t",temp_id)
        if temp_id not in gene_symbol_dict:
            print('weird mapping event with:',temp_id)
        elif temp_hu_symbol != None:
            temp_symbol_list = gene_symbol_dict[temp_id]
            temp_symbol_list.append(temp_hu_symbol)
            gene_symbol_dict[temp_id] = temp_symbol_list
            temp_def_dict = gene_def_dict[temp_id]
            temp_def_dict.append(temp_def)

    return(gene_symbol_dict,gene_def_dict,results)

##############################################################

def pyminer_gprofiler_converter_main(ID_list,
                                     out,
                                     species='hsapiens',
                                     annotations = False):
    ## if we will need to convert to human IDs
    if species != "hsapiens" and not annotations:
        convert = True
    else:
        convert = False


    ###########################
    ## convert id_list to a gene lookup table and write it to file

    ## load in the ID list
    temp_ID_list = read_file(ID_list)
    ID_list = []
    for temp_id in temp_ID_list:
        try:
            int(float(temp_id))
        except:
            ID_list.append(temp_id)
        else:
            ID_list.append(str(int(float(temp_id))))
            #print('found int')


    ## convert to ensg
    symbol_dict, def_dict, annotation_table = convert_to_ensg(ID_list, species)
    write_table(annotation_table,args.out+".tsv")

    for i in range(0,5):
        print(ID_list[i])
        print('\t',symbol_dict[ID_list[i]])

    save_dict((symbol_dict,def_dict),args.out+".pkl")

    ## write the ortholgoues
    if convert:
        human_symbol_dict, human_def_dict, full_orthologue_table = convert_to_human(ID_list, species)
        save_dict((human_symbol_dict, human_def_dict), args.out+".pkl")
        write_table(full_orthologue_table,args.out+".tsv")

    return



if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-ID_list",'-ids', '-IDs','-i', 
        help="a file containing gene ids separated on new lines",
        type = str)
    parser.add_argument("-out", '-outfile', '-o',
        help="the file that we'll write the results to",
        type = str)
    parser.add_argument("-species", '-s',
        help="a gProfiler accepted species code. Dafault = 'hsapiens'",
        type = str,
        default = 'hsapiens')
    parser.add_argument("-annotations",
        help="just get the annotation table",
        action = 'store_true',
        default = False)
    args = parser.parse_args()
    pyminer_gprofiler_converter_main(ID_list = args.ID_list,
                                     out = args.out,
                                     species = args.species,
                                     annotations = args.annotations)
##############################################################




##############################################################
##############################################################


##############################################################
##############################################################










#!/usr/bin/env python3

import fileinput, h5py, argparse, os, sys
import numpy as np

try:
    from pyminer.common_functions import *
except:
    from common_functions import *


def h5_to_tab(infile, rows, cols, out):
	infile = os.path.abspath(infile)
	out = os.path.abspath(out)
	## read the hdf5
	h5f = h5py.File(infile, 'r')
	in_mat = h5f["infile"]
	row_labs = read_file(rows)
	col_labs = read_file(cols)
	## will need to add new line only to non-last lines
	counter_end = len(row_labs)-1
	counter = 0
	## set up the output file
	f = open(out, 'w')
	## write the first line
	f.write('\t'.join(col_labs)+'\n')
	## write the subsequent lines
	for i in range(len(row_labs)):
		if counter % 1000==0:
			print('\tline',counter)
		temp_line = [row_labs[i]]+list(map(str,in_mat[i,:].tolist()))
		temp_line = '\t'.join(temp_line)
		if i != counter_end:
			temp_line+='\n'
		else:
			print("reached the end of the file:",counter)
		f.write(temp_line)
		counter+=1
	h5f.close()
	f.close()
	return()





if __name__ == '__main__':
	parser = argparse.ArgumentParser()
	parser.add_argument('-infile','-i',
		                type=str,
		                help = "this must be a tab delimited file which will be converted to the hdf5 format (output will be created with the same filename with the .hdf5 extention)")
	parser.add_argument('-rows','-ID_list', '-ids',
		                type=str,
		                help = "")
	parser.add_argument('-cols','-columns',
		                type=str,
		                help = "")
	parser.add_argument('-out','-o',
		                type=str,
		                help = "")
	args = parser.parse_args()
	h5_to_tab(args.infile, args.rows, args.cols, args.out)


#!/usr/bin/env python3

##import dependency libraries
import sys,time,glob,os,pickle,fileinput,argparse, random
from time import sleep
from subprocess import Popen
from operator import itemgetter
import gc, fileinput
import numpy as np
#import pandas as pd
##############################################################
## basic function library
def read_file(tempFile,linesOraw='lines',quiet=False):
    if not quiet:
        print('reading',tempFile)
    f=open(tempFile,'r')
    if linesOraw=='lines':
        lines=f.readlines()
        for i in range(0,len(lines)):
            lines[i]=lines[i].strip('\n')
    elif linesOraw=='raw':
        lines=f.read()
    f.close()
    return(lines)

def make_file(contents,path):
    f=open(path,'w')
    if isinstance(contents,list):
        f.writelines(contents)
    elif isinstance(contents,str):
        f.write(contents)
    f.close()

    
def flatten_2D_table(table,delim):
    #print(type(table))
    if str(type(table))=="<class 'numpy.ndarray'>":
        out=[]
        for i in range(0,len(table)):
            out.append([])
            for j in range(0,len(table[i])):
                try:
                    str(table[i][j])
                except:
                    print(table[i][j])
                else:
                    out[i].append(str(table[i][j]))
            out[i]=delim.join(out[i])+'\n'
        return(out)
    else:
        for i in range(0,len(table)):
            for j in range(0,len(table[i])):
                try:
                    str(table[i][j])
                except:
                    print(table[i][j])
                else:
                    table[i][j]=str(table[i][j])
            table[i]=delim.join(table[i])+'\n'
    #print(table[0])
        return(table)

def strip_split(line, delim = '\t'):
    return(line.strip('\n').split(delim))

def make_table(lines,delim):
    for i in range(0,len(lines)):
        lines[i]=lines[i].strip()
        lines[i]=lines[i].split(delim)
        for j in range(0,len(lines[i])):
            try:
                float(lines[i][j])
            except:
                lines[i][j]=lines[i][j].replace('"','')
            else:
                lines[i][j]=float(lines[i][j])
    return(lines)


def get_file_path(in_path):
    in_path = in_path.split('/')
    in_path = in_path[:-1]
    in_path = '/'.join(in_path)
    return(in_path+'/')


def read_table(file, sep='\t'):
    return(make_table(read_file(file,'lines'),sep))
    
def write_table(table, out_file, sep = '\t'):
    make_file(flatten_2D_table(table,sep), out_file)
    

def import_dict(f):
    f=open(f,'rb')
    d=pickle.load(f)
    f.close()
    return(d)

def save_dict(d,path):
    f=open(path,'wb')
    pickle.dump(d,f)
    f.close()

all_calls = []
def cmd(in_message, com=True):
    global all_calls, time, sleep
    if type(in_message)==list:
       in_message = ' '.join(in_message)
    print(in_message)
    all_calls.append(in_message)
    #time.sleep(.25)
    sleep(0.25)
    if com:
       Popen(in_message,shell=True).communicate()
    else:
       Popen(in_message,shell=True)

def check_infile(infile):
	if os.path.isfile(infile):
		return
	else:
		sys.exit(str('could not find '+infile))

def outfile_exists(outfile):
	if os.path.isfile(outfile):
		statinfo = os.stat(outfile)
		if statinfo.st_size!=0:
			return(True)
		else:
			return(False)
	else:
		return(False)
##############################################################

import argparse
parser = argparse.ArgumentParser()

parser.add_argument(
    "-gene_lists",
    help='a comma separated list of files that contain the enriched genes for each cell type.',
    type = str)

parser.add_argument(
    "-gene_table",'-bool_table','-in_table',
    help='A boolean True False table indicating whether a gene should be included in each cell types list.',
    type = str)

parser.add_argument(
    "-organism",'-species','-s',
    help='which organism are we using. (Use the -species_codes to print out a list of all supported species and their codes).',
    type = str,
    default = 'hsapiens')

parser.add_argument(
    "-species_codes",
    help='print out a list of all supported species and their codes.',
    action = 'store_true',
    default = False)

parser.add_argument(
    "-convert_to_human",
    help='if using a non-mouse/human species, it may be useful to convert to human, as these interactions will be more complete. Note that you will still have to provide an -organism argument so that we know what species to convert from.',
    action = 'store_true',
    default = False)

parser.add_argument(
    "-out_dir",'-out','-o',
    help='The output directory to use. If it does not exist, it will be created.',
    type = str)


parser.add_argument(
    "-stringdb_dir",'-sdb',
    help='The directory containing the StringDB action files',
    type = str,
    default = '/usr/local/lib/cell_signals/')

args = parser.parse_args()
##############################################################
##############################################################

organism_action_files = {
    'hsapiens':'9606.protein.actions.v11.0.txt',
    'mmusculus':'10090.protein.actions.v11.0.txt'
}


if args.organism not in organism_action_files:
    args.convert_to_human = True

## check that we have gprofiler installed
try:
    from gprofiler import GProfiler
except:
    sys.exit('please install gprofiler; try: pip3 install gprofiler-official')
else:
    from gprofiler import GProfiler

gp = GProfiler('PyMINEr_'+str(random.randint(0,int(1e6))), want_header = True)

##############################################################
##############################################################
if args.species_codes:
    print("""
Ensembl

    aaegypti — Aedes aegypti
    acarolinensis — Anolis carolinensis
    agambiae — Anopheles gambiae
    amelanoleuca — Ailuropoda melanoleuca
    amexicanus — Astyanax mexicanus
    aplatyrhynchos — Anas platyrhynchos
    btaurus — Bos taurus
    celegans — Caenorhabditis elegans
    cfamiliaris — Canis familiaris
    choffmanni — Choloepus hoffmanni
    cintestinalis — Ciona intestinalis
    cjacchus — Callithrix jacchus
    cporcellus — Cavia porcellus
    csabaeus — Chlorocebus sabaeus
    csavignyi — Ciona savignyi
    dmelanogaster — Drosophila melanogaster
    dnovemcinctus — Dasypus novemcinctus
    dordii — Dipodomys ordii
    drerio — Danio rerio
    ecaballus — Equus caballus
    eeuropaeus — Erinaceus europaeus
    etelfairi — Echinops telfairi
    falbicollis — Ficedula albicollis
    fcatus — Felis catus
    gaculeatus — Gasterosteus aculeatus
    ggallus — Gallus gallus
    ggorilla — Gorilla gorilla
    gmorhua — Gadus morhua
    hsapiens — Homo sapiens
    itridecemlineatus — Ictidomys tridecemlineatus
    lafricana — Loxodonta africana
    lchalumnae — Latimeria chalumnae
    loculatus — Lepisosteus oculatus
    mdomestica — Monodelphis domestica
    meugenii — Macropus eugenii
    mfuro — Mustela putorius furo
    mgallopavo — Meleagris gallopavo
    mlucifugus — Myotis lucifugus
    mmulatta — Macaca mulatta
    mmurinus — Microcebus murinus
    mmusculus — Mus musculus
    nleucogenys — Nomascus leucogenys
    oanatinus — Ornithorhynchus anatinus
    oaries — Ovis aries
    ocuniculus — Oryctolagus cuniculus
    ogarnettii — Otolemur garnettii
    olatipes — Oryzias latipes
    oniloticus — Oreochromis niloticus
    oprinceps — Ochotona princeps
    pabelii — Pongo abelii
    panubis — Papio anubis
    pcapensis — Procavia capensis
    pformosa — Poecilia formosa
    pmarinus — Petromyzon marinus
    psinensis — Pelodiscus sinensis
    ptroglodytes — Pan troglodytes
    pvampyrus — Pteropus vampyrus
    rnorvegicus — Rattus norvegicus
    saraneus — Sorex araneus
    scerevisiae — Saccharomyces cerevisiae
    sharrisii — Sarcophilus harrisii
    sscrofa — Sus scrofa
    tbelangeri — Tupaia belangeri
    tguttata — Taeniopygia guttata
    tnigroviridis — Tetraodon nigroviridis
    trubripes — Takifugu rubripes
    tsyrichta — Tarsius syrichta
    ttruncatus — Tursiops truncatus
    vpacos — Vicugna pacos
    xmaculatus — Xiphophorus maculatus
    xtropicalis — Xenopus tropicalis

Ensembl Genomes Fungi

    aclavatus — Aspergillus clavatus
    aflavus — Aspergillus flavus
    afumigatus — Aspergillus fumigatus
    afumigatusa1163 — Aspergillus fumigatusa1163
    agossypii — Ashbya gossypii
    anidulans — Aspergillus nidulans
    aniger — Aspergillus niger
    aoryzae — Aspergillus oryzae
    aterreus — Aspergillus terreus
    bcinerea — Botrytis cinerea
    bgraminis — Blumeria graminis f. sp. hordei DH14
    cgloeosporioides — Colletotrichum gloeosporioides
    cgraminicola — Colletotrichum graminicola
    chigginsianum — Colletotrichum higginsianum
    cneoformans — Cryptococcus neoformans
    corbiculare — Colletotrichum orbiculare
    dseptosporum — Dothistroma septosporum
    fculmorum — Fusarium culmorum UK99
    ffujikuroi — Fusarium fujikuroi
    fgraminearum — Fusarium graminearum
    foxysporum — Fusarium oxysporum
    fpseudograminearum — Fusarium pseudograminearum
    fsolani — Fusarium solani
    fverticillioides — Fusarium verticillioides
    ggraminis — Gaeumannomyces graminis
    kpastoris — Komagataella pastoris
    lmaculans — Leptosphaeria maculans
    mlaricipopulina — Melampsora larici-populina 98AG31
    moryzae — Magnaporthe oryzae
    mpoae — Magnaporthe poae
    mviolaceum — Microbotryum violaceum p1A1 Lamole
    ncrassa — Neurospora crassa
    nfischeri — Neosartorya fischeri
    pgraminis — Puccinia graminis
    pgraminisug99 — Puccinia graminis Ug99
    pnodorum — Phaeosphaeria nodorum
    pstriiformis — Puccinia striiformis f. sp. tritici PST-130 str. Race 130
    pteres — Pyrenophora teres f. teres 0-1
    ptriticina — Puccinia triticina
    ptriticirepentis — Pyrenophora tritici-repentis Pt-1C-BFP
    scerevisiae — Saccharomyces cerevisiae
    scryophilus — Schizosaccharomyces cryophilus
    sjaponicus — Schizosaccharomyces japonicus
    soctosporus — Schizosaccharomyces octosporus
    spombe — Schizosaccharomyces pombe
    sreilianum — Sporisorium reilianum SRZ2
    ssclerotiorum — Sclerotinia sclerotiorum
    tmelanosporum — Tuber melanosporum
    treesei — Trichoderma reesei
    tvirens — Trichoderma virens
    umaydis — Ustilago maydis
    vdahliae — Verticillium dahliae
    vdahliaejr2 — Verticillium dahliae JR2
    ylipolytica — Yarrowia lipolytica
    ztritici — Zymoseptoria tritici

Ensembl Genomes Metazoa

    aaegypti — Aedes aegypti
    acephalotes — Atta cephalotes
    adarlingi — Anopheles darlingi
    agambiae — Anopheles gambiae
    aglabripennis — Anoplophora glabripennis
    amellifera — Apis mellifera
    apisum — Acyrthosiphon pisum
    aqueenslandica — Amphimedon queenslandica
    avaga — Adineta vaga
    bantarctica — Belgica antarctica
    bimpatiens — Bombus impatiens
    bmalayi — Brugia malayi
    bmori — Bombyx mori
    cbrenneri — Caenorhabditis brenneri
    cbriggsae — Caenorhabditis briggsae
    celegans — Caenorhabditis elegans
    cgigas — Crassostrea gigas
    cjaponica — Caenorhabditis japonica
    cquinquefasciatus — Culex quinquefasciatus
    cremanei — Caenorhabditis remanei
    cteleta — Capitella teleta
    dananassae — Drosophila ananassae
    derecta — Drosophila erecta
    dgrimshawi — Drosophila grimshawi
    dmelanogaster — Drosophila melanogaster
    dmojavensis — Drosophila mojavensis
    dpersimilis — Drosophila persimilis
    dplexippus — Danaus plexippus
    dponderosae — Dendroctonus ponderosae
    dpseudoobscura — Drosophila pseudoobscura
    dpulex — Daphnia pulex
    dsechellia — Drosophila sechellia
    dsimulans — Drosophila simulans
    dvirilis — Drosophila virilis
    dwillistoni — Drosophila willistoni
    dyakuba — Drosophila yakuba
    hmelpomene — Heliconius melpomene
    hrobusta — Helobdella robusta
    iscapularis — Ixodes scapularis
    lanatina — Lingula anatina
    lcuprina — Lucilia cuprina
    lgigantea — Lottia gigantea
    lloa — Loa loa
    lsalmonis — Lepeophtheirus salmonis
    mcinxia — Melitaea cinxia
    mdestructor — Mayetiola destructor
    mleidyi — Mnemiopsis leidyi
    mscalaris — Megaselia scalaris
    nvectensis — Nematostella vectensis
    nvitripennis — Nasonia vitripennis
    obimaculoides — Octopus bimaculoides
    ovolvulus — Onchocerca volvulus
    phumanus — Pediculus humanus
    ppacificus — Pristionchus pacificus
    rprolixus — Rhodnius prolixus
    sinvicta — Solenopsis invicta
    smansoni — Schistosoma mansoni
    smaritima — Strigamia maritima
    smimosarum — Stegodyphus mimosarum
    spurpuratus — Strongylocentrotus purpuratus
    sratti — Strongyloides ratti
    sscabiei — Sarcoptes scabiei
    tadhaerens — Trichoplax adhaerens
    tcastaneum — Tribolium castaneum
    tkitauei — Thelohanellus kitauei
    tspiralis — Trichinella spiralis
    turticae — Tetranychus urticae
    znevadensis — Zootermopsis nevadensis

Ensembl Genomes Plants

    alyrata — Arabidopsis lyrata
    atauschii — Aegilops tauschii
    athaliana — Arabidopsis thaliana
    atrichopoda — Amborella trichopoda
    bdistachyon — Brachypodium distachyon
    bnapus — Brassica napus
    boleracea — Brassica oleracea
    brapa — Brassica rapa
    bvulgaris — Beta vulgaris subsp. vulgaris
    ccrispus — Chondrus crispus
    cmerolae — Cyanidioschyzon merolae
    creinhardtii — Chlamydomonas reinhardtii
    gmax — Glycine max
    gsulphuraria — Galdieria sulphuraria
    hvulgare — Hordeum vulgare
    lperrieri — Leersia perrieri
    macuminata — Musa acuminata
    mtruncatula — Medicago truncatula
    obarthii — Oryza barthii
    obrachyantha — Oryza brachyantha
    oglaberrima — Oryza glaberrima
    oglumaepatula — Oryza glumaepatula
    oindica — Oryza sativa indica
    olongistaminata — Oryza longistaminata
    olucimarinus — Ostreococcus lucimarinus
    omeridionalis — Oryza meridionalis
    onivara — Oryza nivara
    opunctata — Oryza punctata
    orufipogon — Oryza rufipogon
    osativa — Oryza sativa Japonica
    ppatens — Physcomitrella patens
    ppersica — Prunus persica
    ptrichocarpa — Populus trichocarpa
    sbicolor — Sorghum bicolor
    sitalica — Setaria italica
    slycopersicum — Solanum lycopersicum
    smoellendorffii — Selaginella moellendorffii
    stuberosum — Solanum tuberosum
    taestivum — Triticum aestivum
    tcacao — Theobroma cacao
    tpratense — Trifolium pratense
    turartu — Triticum urartu
    vvinifera — Vitis vinifera
    zmays — Zea mays
    """)
    sys.exit()

##############################################################
##############################################################



def get_name_from_path(in_path):
    temp_name = in_path.split('/')
    temp_name = temp_name[-1]
    temp_name = temp_name[:-4]
    return(temp_name)

def get_gene_lists_from_files(in_files):
    all_gene_list_files = in_files.split(',')
    cell_type_ids = []
    for gene_list in all_gene_list_files:
        cell_type_ids.append(get_name_from_path(gene_list))

    ## get the gene lists for each cell type
    gene_lists = []
    gene_ids = []
    for cell in all_gene_list_files:
        temp_genes = read_file(cell)
        gene_lists.append(temp_genes)
        gene_ids += temp_genes
    gene_ids = list(set(gene_ids))
    return(cell_type_ids, gene_ids, gene_lists)

def get_gene_lists_from_bool_table(in_file):
    """
    here, we take in the boolean table with columns corresponding
    to the cell type names, with the body of the table consisting of
    True or False. This will be directly translated into a boolean numpy array
    so syntax is important in this array.
    """
    bool_table = read_table(in_file)
    cell_type_ids = bool_table[0][1:]
    bool_table = np.array(bool_table)
    gene_ids = list(set(bool_table[1:,0].tolist()))
    gene_lists = []
    ## set up the empty lists
    for i in range(1, np.shape(bool_table)[1]):
        gene_lists.append([])
    ## populate the 
    for i in range(1,np.shape(bool_table)[0]):
        ## go through each row in the table (barring the )
        for j in range(1, np.shape(bool_table)[1]):
            if bool_table[i,j]=='True':
                gene_lists[j-1].append(bool_table[i,0])
    
    ## make them all upper case
    for i in range(0,len(gene_ids)):
        gene_ids[i]=str(gene_ids[i]).upper()

    for i in range(0,len(gene_lists)):
        for j in range(0,len(gene_lists[i])):
            gene_lists[i][j]=str(gene_lists[i][j]).upper()

    return(cell_type_ids, gene_ids, gene_lists)


def convert_to_ensp(in_genes, gene_lists):
    global args, gp
    temp_lookup={}
    gene_symbol_dict = {}
    for g in in_genes:
        temp_lookup[g] = []
        


    results = gp.gconvert(in_genes, 
        organism = args.organism, target='ENSP')


    for r in range(1,len(results)):
        ## the original ID
        g = results[r][1]
        ## the current orthologue list
        try:
            temp_lookup[g]
        except:
            pass
            # print('weird mapping event:')
            # print(results[r])
        else:
            if results[r][4] != None:
                gene_symbol_dict[results[r][3]]=results[r][4]
            if results[r][3] != None:
                temp_list = temp_lookup[g]
                temp_list += [results[r][3]]
                temp_lookup[g]=temp_list

    gene_lists = convert_list_of_lists_by_dict(gene_lists,temp_lookup)
    new_gene_ids = []
    for k in list(temp_lookup.keys()):
        new_gene_ids += temp_lookup[k]
    new_gene_ids= list(set(new_gene_ids))

    for gene in new_gene_ids:
        if gene not in gene_symbol_dict:
            gene_symbol_dict[gene]="None"


    return(new_gene_ids, gene_lists, gene_symbol_dict)


def get_orthologue_dicts(in_genes,source_organism, target_organism, orthologues = True, ensp_convert = False):
    global args, gp
    temp_org_to_human = {}

    ## each gene will get a list for its orthologues
    for g in in_genes:
        temp_org_to_human[g] = []

    results = gp.gorth(in_genes, 
        source_organism = source_organism, 
        target_organism=target_organism,
        numeric_ns=True)


    for r in range(1,len(results)):

        ## the original ID
        g = results[r][1]
        ## the current orthologue list
        try:
            temp_org_to_human[g]
        except:
            pass
            # print('weird mapping event:')
            # print(results[r])
        else:
            if results[r][4] != None:
                temp_list = temp_org_to_human[g]
                temp_list += [results[r][4]]
                temp_org_to_human[g]=temp_list

    human_to_temp_org = {}
    return(temp_org_to_human,human_to_temp_org)

def convert_list_of_lists_by_dict(gene_lists,temp_org_to_human):
    ## set up the new gene list
    new_gene_lists = []
    for i in range(len(gene_lists)):
        new_gene_lists.append([])
    ## convert the original gene lists to the IDs
    for i in range(len(gene_lists)):
        for j in range(len(gene_lists[i])):
            new_gene_lists[i]+=temp_org_to_human[gene_lists[i][j]]
    return(new_gene_lists)


def update_converted_genes(temp_org_to_human, gene_ids, gene_lists):
    new_gene_ids = []
    for k in list(temp_org_to_human.keys()):
        new_gene_ids += temp_org_to_human[k]
    new_gene_ids= list(set(new_gene_ids))
    if None in new_gene_ids:
        new_gene_ids.pop(None)

    new_gene_lists=convert_list_of_lists_by_dict(gene_lists,temp_org_to_human)

    return(new_gene_ids, new_gene_lists)

def get_receptors_and_secreted(in_genes):
    global gp

    secreted_go = ['GO:0005615','GO:0005576','GO:0044421']
    receptor_go = ["GO:0009897","GO:0031232","GO:0031233","GO:0071575"," GO:0098591","GO:0031362","GO:0098567","GO:0009986","GO:0005886","GO:0042923","GO:0016021"]#,'GO:0005904']

    results = gp.gconvert(gene_ids, 
        organism = args.organism, target='GO')


    # print(results[0])
    # print(results[1])
    # print(results[2])
    # print(results[3])

    all_go_terms = []
    receptor_list = []
    secreted_list = []
    for i in range(1,len(results)):
        if results[i][3] not in all_go_terms:
            all_go_terms.append(results[i][3])
        if results[i][3] in receptor_go:
            receptor_list.append(results[i][1])
        if results[i][3] in secreted_go:
            secreted_list.append(results[i][1])
    receptor_list = list(set(receptor_list))
    secreted_list = list(set(secreted_list))
    return(receptor_list, secreted_list)

def filter_enriched_lists_for_subcellular_localization(receptor_list, secreted_list, gene_lists):
    receptor_gene_lists = []
    secreted_gene_lists = []
    for g in range(len(gene_lists)):
        receptor_gene_lists.append([])
        secreted_gene_lists.append([])

    for i in range(0,len(gene_lists)):
        for j in range(0,len(gene_lists[i])):
            if gene_lists[i][j] in receptor_list:
                receptor_gene_lists[i].append(gene_lists[i][j])
            if gene_lists[i][j] in secreted_list:
                secreted_gene_lists[i].append(gene_lists[i][j])
    return(receptor_gene_lists,secreted_gene_lists)

def clean_names(in_vect):
    name1 = in_vect[0].split('.')
    name2 = in_vect[1].split('.')
    name1 = name1[-1]
    name2 = name2[-1]
    in_vect[0] = name1
    in_vect[1] = name2
    return(in_vect)

def subset_action_table(receptor_or_secreted_list,db_file):
    print('this may take a minute....')
    ## construct a dictionary for the receptor or secreted list
    receptor_or_secreted_dict = {j:i for i, j in enumerate(receptor_or_secreted_list)}
    print(len(list(receptor_or_secreted_dict.keys())), 'receptor_or_secreted_dict')
    all_ids = []
    line_count = 0
    out_table = []
    activation_dict = {}
    inhibition_dict = {}
    other_dict = {}
    first = True
    for line in fileinput.input(db_file):
        line_count+=1
        if line_count % 1e5 == 0:
            print('\t',line_count,'\t',len(out_table))
        temp_line = strip_split(line)
        if first:
            first = False
            #out_table.append(temp_line) 
        else:
            temp_line = clean_names(temp_line)


            ## check if the two genes are in the receptor or secreted list
            # holy shit this way of checking if something is in a dictionary is so much
            # faster than checking if they're in a list... good to know...
            # I'm talking like, orders of magnitude faster...
            # I've wasted so much time waiting on this step. Now it takes seconds o...
            cont1 = True
            try:
                receptor_or_secreted_dict[temp_line[0]]
            except:
                cont1=False
            if cont1:
                all_ids.append(temp_line[0])

            cont2 = True
            try:
                receptor_or_secreted_dict[temp_line[1]]
            except:
                cont2=False
            if cont2:
                all_ids.append(temp_line[1])

            if cont1 and cont2:
            #if temp_line[0] in receptor_or_secreted_list and temp_line[1] in receptor_or_secreted_list:
                temp_id = temp_line[0]+":"+temp_line[1]
                if temp_line[2] == 'binding': 
                    out_table.append(temp_line)
                elif temp_line[2] == 'activation':
                    activation_dict[temp_id] = True
                elif temp_line[2] == 'inhibition':
                    inhibition_dict[temp_id] = True
                else:
                    if temp_id not in other_dict:
                        other_dict[temp_id]=temp_line[2]
                    else:
                        if other_dict[temp_id] != 'None':
                            temp_thingy = other_dict[temp_id]
                            temp_thingy = ','+temp_line[2]
                        if temp_thingy[0]==',':
                            temp_thingy=temp_thingy[1:]
                        other_dict[temp_id] = temp_thingy

    fileinput.close()
    all_ids = list(set(all_ids))
    return(out_table, activation_dict, inhibition_dict, other_dict,all_ids)

def annotate_interactions_for_activation_and_inhibition(action_table, activation_dict, inhibition_dict, other_dict):
    print('annotating the interactions for directionality and type of action....')
    action_table[0] = action_table[0][:11]+['activation/inhibition','other_action']+action_table[0][11:]
    # print(action_table[0])
    for i in range(1,len(action_table)):
    #     for l in enumerate(action_table[i]):
    #         print(l)
        temp_id = str(action_table[i][9])+":"+str(action_table[i][10])
        activation_term = 'None'
        other_action_term = 'None'
        if temp_id in activation_dict:
            activation_term = 'activation'
        if temp_id in inhibition_dict:
            activation_term = 'inhibition'
        if temp_id in other_dict:
            other_action_term = other_dict[temp_id]
        action_table[i] = action_table[i][:11]+[activation_term,other_action_term]+action_table[i][11:]
    return(action_table)



def gene_lists_to_symbol_lists(gene_lists, gene_symbol_dict):
    gene_symbol_lists = []
    for c in range(len(gene_lists)):
        gene_symbol_lists.append([])
        for gene in gene_lists[c]:
            temp_symbol = gene_symbol_dict[gene]
            if temp_symbol not in gene_symbol_lists[c]:
                gene_symbol_lists[c].append(temp_symbol)
    return(gene_symbol_lists)

def gene_symbol_lists_to_dicts(gene_symbol_lists):
    gene_symbol_list_dicts = []
    for i in range(0,len(gene_symbol_lists)):
        gene_symbol_list_dicts.append({j:w for w, j in enumerate(gene_symbol_lists[i])})
    return(gene_symbol_list_dicts)
##############################################################
##############################################################




##############################################################
##############################################################
print('\n')
if args.gene_table!=None:
    cell_type_ids, gene_ids, gene_lists = get_gene_lists_from_bool_table(args.gene_table)


if args.gene_lists!=None:
    cell_type_ids, gene_ids, gene_lists = get_gene_lists_from_files(args.gene_lists)

print(len(gene_ids))

#################
## print out some basic summaries on the input data
print(len(cell_type_ids),'cell types found')
print(cell_type_ids[:min([4,len(cell_type_ids)])],'...\n')

print(len(gene_ids),'genes found')
print(gene_ids[:min([4,len(gene_ids)])],'...\n')

## make a hash table to look up the index from a cell type name
cell_type_idx_dict = {key:value for value, key in enumerate(cell_type_ids)}
# for c in cell_type_ids:
#     print(c, cell_type_idx_dict[c])
#################

#################
## if we need to convert over to human ids, this is the first thing we'll need to do
if args.convert_to_human:
    print('converting to human')
    temp_org_to_human, human_to_temp_org = get_orthologue_dicts(gene_ids, args.organism, 'hsapiens')
    #get_orthologue_dicts(gene_ids[:10], args.organism, 'hsapiens')

    ## update the gene_ids and gene_lists to reflect the conversion to human
    gene_ids, gene_lists = update_converted_genes(temp_org_to_human, gene_ids, gene_lists)
    args.organism = 'hsapiens'
    print('\n\nfinished converting IDs\n\n')

    print('gene IDs now look like:')
    print(gene_ids[:4],'...\n')
    print('enrichment lists now look like:')
    print('\t',cell_type_ids[0])
    print('\t\t',gene_lists[0][:3])
    print('\t',cell_type_ids[1],'...')
    print('\t\t',gene_lists[1][:3],'...\n\n')



#################
print('Converting IDs to ENSP format.')
print('This is needed for cross-referencing the StringDB database later\n')


# for i in range(0,len(gene_lists)):
#     l=gene_lists[i]
#     if "ENSG00000139874" in l:
#         print('found "ENSG00000139874" in',i)

gene_ids, gene_lists, gene_symbol_dict = convert_to_ensp(gene_ids, gene_lists)
gene_symbol_lists = gene_lists_to_symbol_lists(gene_lists, gene_symbol_dict)
gene_symbol_list_dicts = gene_symbol_lists_to_dicts(gene_symbol_lists)


# for i in range(0,len(gene_lists)):
#     l=gene_lists[i]
#     if "ENSP00000267377" in l:
#         print('found "ENSP00000267377" in',i)



print('finished converting to ENSPs\n')
print('gene IDs now look like:')
print(gene_ids[:4],'...\n')
print('enrichment lists now look like:')
print('\t',cell_type_ids[0])
print('\t\t',gene_lists[0][:3])
#print('\t',cell_type_ids[1],'...')
#print('\t\t',gene_lists[1][:3],'...\n\n')


symbols = list(gene_symbol_dict.keys())
print(len(symbols))
for i in range(0,5):
    print(symbols[i], gene_symbol_dict[symbols[i]])
#sys.exit()

#####################################################################################
#####################################################################################
###################   go term filter for subcellular localization   #################
#####################################################################################
#####################################################################################

print('getting subcellular localizations\n')
receptor_list, secreted_list = get_receptors_and_secreted(gene_ids)


print('found',len(receptor_list),'plasma membrane')
print('found',len(secreted_list),'extracellular')

# print("in receptor list?:","ENSP00000267377" in receptor_list)
# print("in secreted list?:","ENSP00000267377" in secreted_list)

## if there are any extracellular genes in the plasma membrane gene lists, pop them out
## we only want loose ligands in the secreted list 
for r in receptor_list:
    # print(r)
    # print(secreted_list[:10])
    if r in secreted_list:
        print('removing',r,gene_symbol_dict[r],"from secreted, because it's also annotated as being in the plasma membrane")
        secreted_list.pop(secreted_list.index(r))

# for r in secreted_list:
#     # print(r)
#     # print(secreted_list[:10])
#     if r in receptor_list:
#         print('removing',r,gene_symbol_dict[r],"from receptor, because it's also annotated as being in the secreted")
#         receptor_list.pop(receptor_list.index(r))


print('found',len(receptor_list),'plasma membrane')
print('found',len(secreted_list),'extracellular')
receptor_or_secreted_list = list(set(receptor_list + secreted_list))

print(len(receptor_or_secreted_list), 'receptors and secreted genes found')

# if "ENSP00000267377" in receptor_or_secreted_list:
#     print("found ENSP00000267377 in receptor_or_secreted_list")
# else:
#     print("Couldn't find ENSP00000267377 in receptor or secreted!")
#     sys.exit()

## set up a dictionary with the info needed about each gene for writing to file later
gene_detail_dict = {}
for g in gene_ids:
    if g in receptor_list:
        gene_detail_dict[g]=[g,gene_symbol_dict[g],'plasma_membrane']
    if g in secreted_list:
        gene_detail_dict[g]=[g,gene_symbol_dict[g],'extracellular']

## subset the enriched gene lists for each subcellular localization
receptor_gene_lists, secreted_gene_lists = filter_enriched_lists_for_subcellular_localization(receptor_list, secreted_list, gene_lists)

## print out some summaries
print('plasma membrane:')
for i in range(0,len(cell_type_ids)):
    temp_list = receptor_gene_lists[i]
    temp_list_len = len(receptor_gene_lists[i])
    print('\t',cell_type_ids[i],temp_list_len)
    if temp_list_len!=0:
        print('\t\t',receptor_gene_lists[i][0:min([temp_list_len,3])],'...')

print('\nextracellular:')
for i in range(0,len(cell_type_ids)):
    temp_list = secreted_gene_lists[i]
    temp_list_len = len(secreted_gene_lists[i])
    print('\t',cell_type_ids[i],temp_list_len)
    if temp_list_len!=0:
        print('\t\t',secreted_gene_lists[i][0:min([temp_list_len,3])],'...')


# for i in range(0,len(secreted_gene_lists)):
#     l=secreted_gene_lists[i]
#     if "ENSP00000267377" in l:
#         print('found "ENSP00000267377" in secreted_gene_lists',i)


# for i in range(0,len(receptor_gene_lists)):
#     l=receptor_gene_lists[i]
#     if "ENSP00000267377" in l:
#         print('found "ENSP00000267377" in receptor_gene_lists',i)

#####################################################################################
#####################################################################################
###################   cross reference all of the receptor and secreted   ############
###################       lists with the StringDB action database        ############
#####################################################################################
#####################################################################################

if args.organism in organism_action_files:
    db_file = args.stringdb_dir+organism_action_files[args.organism]

#########
## first we'll load into memory only the interactions that stand
## a chance of being pertinent to our dataset
print('\n\nloading in the interaction table')
action_table, activation_dict, inhibition_dict, other_dict, all_included_bg = subset_action_table(receptor_or_secreted_list,db_file)
print(len(action_table)-1,'interactions among the genes remaining')

# for i in range(0,len(action_table)):
#     if "ENSP00000267377" in action_table[i]:
#         print("found ENSP00000267377 in action_table:",action_table[i])


#########
## convert that to an action dictionary so that we can look up the details quickly later
# first make a unique key for the interaction lists
action_dict = {}
symbol_interaction_dict = {}
for line in action_table:
    temp_id = line[0]+":"+line[1]
    #print(temp_id)
#    if line[0] in gene_symbol_dict and line[1] in gene_symbol_dict:
    temp_symbol_id = gene_symbol_dict[line[0]]+":"+gene_symbol_dict[line[1]]
# else:
#     temp_symbol_id = None
    # print(temp_id,temp_symbol_id)
    if temp_symbol_id not in symbol_interaction_dict:
        symbol_interaction_dict[temp_symbol_id]=True
    if temp_id not in action_dict:
        action_dict[temp_id] = [line]
    else:
        temp_list = action_dict[temp_id]
        temp_list.append(line)
        action_dict[temp_id] = temp_list



def get_empty_interaction_lists():
    global gene_lists
    interaction_lists = []
    for i in range(0,len(gene_lists)):
        interaction_lists.append([])
        for j in range(0,len(gene_lists)):
            interaction_lists[i].append([])
    return(interaction_lists)

interaction_lists = get_empty_interaction_lists()



interaction_title_line = [["autocrine_paracrine",
"cell_type_1","gene_1","gene_1_symbol","gene_1_location",
"cell_type_2","gene_2","gene_2_symbol","gene_2_location",
"gene1","gene2","action","inhibition","directional","is_direction","score"]]
out_summary_table = [["cell_type_1","cell_type_2","count"]]
out_table = interaction_title_line[:]


## make the cell type enrichment dicts
gene_list_dict = []
for i in range(0,len(gene_lists)):
    temp_gene_dict = {j:i for i, j in enumerate(gene_lists[i])}
    gene_list_dict.append(temp_gene_dict)


def get_cell_types_gene_enriched_in(gene):
    global gene_list_dict, gene_lists, cell_type_ids
    gene_cell_types = []
    for i in range(0,len(cell_type_ids)):
        cur_cell_type = True
        try:
            gene_list_dict[i][gene]
        except:
            cur_cell_type = False
        if cur_cell_type:
            gene_cell_types.append(cell_type_ids[i])
    # ## just for trouble shooting
    # if gene =="ENSP00000357461" or gene == "ENSP00000383938":
    #     print('\n\n',gene_cell_types,'\n\n')
    return(gene_cell_types)



## go through the action_dict & find what cell types are enriched for each of the genes
cell_type_interaction_count = np.zeros((len(cell_type_ids),len(cell_type_ids)))
all_interactions = list(action_dict.keys())

# for interaction in all_interactions:
#     if "ENSP00000267377" in interaction:
#         print("found ENSP00000267377 in all_interactions:",interaction)

print('cateloguing which cell types have the interactions')
for interaction_idx in range(0,len(all_interactions)):
    if interaction_idx %500 == 0:
        print('\t',interaction_idx,'\t',interaction_idx/len(all_interactions))
    interaction = all_interactions[interaction_idx]
    genes = interaction.split(':')
    gene1 = genes[0]
    gene2 = genes[1]
    gene1_symbol = gene_symbol_dict[gene1]
    gene2_symbol = gene_symbol_dict[gene2]
    ## go through all of the cell types and find which cell types are enriched for the genes
    gene1_cell_types = get_cell_types_gene_enriched_in(gene1)
    gene2_cell_types = get_cell_types_gene_enriched_in(gene2)



    ## write the output for each of the combinations of cell types
    for cell_type_1 in gene1_cell_types:
        for cell_type_2 in gene2_cell_types:
            if cell_type_1 == cell_type_2:
                out_line =['autocrine']
            else:
                out_line = ['paracrine']
            cell_type_1_idx = cell_type_idx_dict[cell_type_1]
            cell_type_2_idx = cell_type_idx_dict[cell_type_2]
            interaction_lists[cell_type_1_idx][cell_type_2_idx].append(gene1_symbol)
            #if gene2_symbol not in interaction_lists[i][j]:
            interaction_lists[cell_type_1_idx][cell_type_2_idx].append(gene2_symbol)
            out_line += [cell_type_1]+gene_detail_dict[gene1]
            out_line += [cell_type_2]+gene_detail_dict[gene2]
            for line in action_dict[gene1+":"+gene2]:
                out_table.append(out_line+line)

            cell_type_interaction_count[cell_type_1_idx,cell_type_2_idx]+=1

for i in range(0,len(cell_type_ids)):
    cell_type_1 = cell_type_ids[i]
    for j in range(i,len(cell_type_ids)):
        cell_type_2 = cell_type_ids[j]

        temp_count = cell_type_interaction_count[i,j]

        out_summary_table.append([cell_type_1,cell_type_2,temp_count])
## will need to catalogue the number remaining
#out_summary_table.append([cell_type_1,cell_type_2,temp_count])



if False:
    ## go through all the receptor receptor pairs
    for i in range(0,len(gene_lists)):
        gene_list_1 = gene_lists[i]
        cell_type_1 = cell_type_ids[i]
        print(cell_type_1)
        for j in range(i,len(gene_lists)):
            cell_type_2 = cell_type_ids[j]
            gene_list_2 = gene_lists[j]
            print('\t',cell_type_2)
            temp_count = 0
            for gene1 in gene_list_1:
                gene1_symbol = gene_symbol_dict[gene1]
                for gene2 in gene_list_2:
                    gene2_symbol = gene_symbol_dict[gene2]
                    out_line = None

                    g1_g2=True
                    g2_g1=True
                    try:
                        action_dict[gene1+":"+gene2][0]
                    except:
                        g1_g2=False
                    try:
                        action_dict[gene2+":"+gene1][0]
                    except:
                        g2_g1=False


                    if g1_g2:#ene1+":"+gene2 in action_dict:
                        #if action_dict[gene1+":"+gene2]!=[]:
                            
                        #if gene1_symbol not in interaction_lists[i][j]:
                        interaction_lists[i][j].append(gene1_symbol)
                        #if gene2_symbol not in interaction_lists[i][j]:
                        interaction_lists[i][j].append(gene2_symbol)
                        if cell_type_1 == cell_type_2:
                            out_line = ['autocrine']
                        else:
                            out_line = ['paracrine']
                        out_line += [cell_type_1]+gene_detail_dict[gene1]
                        out_line += [cell_type_2]+gene_detail_dict[gene2]
                        temp_count+=1
                        for line in action_dict[gene1+":"+gene2]:
                            out_table.append(out_line+line)
                                                
                    if g2_g1:#ene2+":"+gene1 in action_dict:
                        #if action_dict[gene2+":"+gene1]!=[]:
                        #if gene1_symbol not in interaction_lists[i][j]:
                        interaction_lists[i][j].append(gene1_symbol)
                        #if gene2_symbol not in interaction_lists[i][j]:
                        interaction_lists[i][j].append(gene2_symbol)
                        if cell_type_1 == cell_type_2:
                            out_line = ['autocrine']
                        else:
                            out_line = ['paracrine']
                        out_line += [cell_type_2]+gene_detail_dict[gene2]
                        out_line += [cell_type_1]+gene_detail_dict[gene1]
                        temp_count+=1
                        for line in action_dict[gene2+":"+gene1]:
                            out_table.append(out_line+line)
                            #print(out_table[-1])
                            
            print('\t\t',temp_count)
            out_summary_table.append([cell_type_1,cell_type_2,temp_count])

## annotate the action table for the mechanism of interaction
out_table = annotate_interactions_for_activation_and_inhibition(out_table, activation_dict, inhibition_dict, other_dict)

## remove the duplicates in the interaction_lists:
for i in range(0,len(interaction_lists)):
    for j in range(0,len(interaction_lists[i])):
        interaction_lists[i][j]=list(set(interaction_lists[i][j]))


out_summary_table_copy = out_summary_table[:]
out_table_copy = out_table[:]



if not os.path.isdir(args.out_dir):
    cmd('mkdir '+args.out_dir)

write_table(out_table_copy,args.out_dir+'/all_cell_type_specific_interactions.txt')
write_table(out_summary_table_copy,args.out_dir+'/all_cell_cell_interaction_summary.txt')


##########################################################
##########################################################
## get gprofiler results from each of the interaction lists
def key_in_dict(key, in_dict):
    in_it = True
    try:
        in_dict[key]
    except:
        in_it = False
    return(in_it)


def get_gprofiler_on_interaction_lists(interaction_lists, bg = None):
    global args, gp, time
    print('getting the pathway analysis for each cell type interactions...')

    output_gprofiler = [['cell_type_1','cell_type_2']]
    for i in range(0,len(interaction_lists)):
        ## wait for a random amount of time so that we don't overload the server
        random_time_lag = np.random.randint(low=25,high=100)/100
        sleep(random_time_lag)
        for j in range(i,len(interaction_lists)):
            print('\t',cell_type_ids[i],cell_type_ids[j],":",len(interaction_lists[i][j]),'genes')
            if bg == None:
                temp_gprofile_results = gp.gprofile(interaction_lists[i][j], organism = args.organism)
            else:
                temp_gprofile_results = gp.gprofile(interaction_lists[i][j], custom_bg = bg, organism = args.organism)
            if i==0 and j == 0 and temp_gprofile_results!=[]:
                output_gprofiler[0] += temp_gprofile_results[0]
            for q in range(1,len(temp_gprofile_results)):
                output_gprofiler.append([cell_type_ids[i],cell_type_ids[j]]+temp_gprofile_results[q])
    return(output_gprofiler)

def see_if_symbols_interact(symbol1,symbol2):
    global symbol_interaction_dict
    temp_key = str(symbol1)+":"+str(symbol2)
    if key_in_dict(temp_key,symbol_interaction_dict):
        return(True)
    temp_key = str(symbol2)+":"+str(symbol1)
    if key_in_dict(temp_key,symbol_interaction_dict):
        return(True)
    return(False)


def filter_and_annoatate_genes(in_gene_list, cell_type_1, cell_type_2, test_time=False):
    global symbol_interaction_dict, gene_symbol_lists, cell_type_idx_dict, gene_symbol_list_dicts
    if cell_type_1 != cell_type_2:
        paracrine = True
    else:
        paracrine = False
    in_gene_list = in_gene_list.split(',')
    if len(in_gene_list)==1:
        return("None")
    cell_type_1_idx = cell_type_idx_dict[cell_type_1]
    cell_type_2_idx = cell_type_idx_dict[cell_type_2]
    #cell_type_enriched_ids_1=gene_symbol_lists[cell_type_1_idx]
    #cell_type_enriched_ids_2=gene_symbol_lists[cell_type_2_idx]
    cell_type_enriched_dict_1=gene_symbol_list_dicts[cell_type_1_idx]
    cell_type_enriched_dict_2=gene_symbol_list_dicts[cell_type_2_idx]
    
    ## first make all fully crossed interaction_pairs
    if test_time:
        print('making pairs')
        start = time()
    fully_crossed_possible_pairs = []
    for i in range(0,len(in_gene_list)):
        for j in range(i,len(in_gene_list)):
            #if in_gene_list[j]+":"+in_gene_list[i] not in fully_crossed_possible_pairs:
            fully_crossed_possible_pairs.append(in_gene_list[i]+":"+in_gene_list[j])
            #fully_crossed_possible_pairs.append(in_gene_list[j]+":"+in_gene_list[i])
    fully_crossed_possible_pairs = list(set(fully_crossed_possible_pairs))
    filtered_pairs = []
    if test_time:
        print('\t',time()-start)
        print('\tlooking at fully_crossed_possible_pairs in interaction dict')
        start = time()
    for i in range(0,len(fully_crossed_possible_pairs)):
        cont = True
        try:
            symbol_interaction_dict[fully_crossed_possible_pairs[i]]
        except:
            cont = False
        if cont:#fully_crossed_possible_pairs[i] in symbol_interaction_dict:
            filtered_pairs.append(fully_crossed_possible_pairs[i])
    #print(filtered_pairs)
    if test_time:
        print('\t\t',time()-start)
        print('\tannotating source')
        start = time()
    final_output_list = []
    for i in range(0,len(filtered_pairs)):
        ## format of this will be cell_type_1:gene1;cell_type2:gene2,... for each interaction contained within the genes in this pathway
        temp_interaction = filtered_pairs[i]
        temp_interaction = temp_interaction.split(':')
        gene_source_1 = ''
        gene_source_2 = ''
        ## figure out which cell type is producing each within the pair 
        if key_in_dict(temp_interaction[0], cell_type_enriched_dict_1):
            #print(temp_interaction[0],"found in",cell_type_1,cell_type_enriched_ids_1)
            gene_source_1+=cell_type_1
        if key_in_dict(temp_interaction[0], cell_type_enriched_dict_2):
            #temp_interaction[0] in cell_type_enriched_ids_2:
            #print(temp_interaction[0],"found in",cell_type_2,cell_type_enriched_ids_2)
            if gene_source_1!='':
                gene_source_1+='&'
            gene_source_1+=cell_type_2
        #print(gene_source_1)
        if key_in_dict(temp_interaction[1], cell_type_enriched_dict_1):
        #if temp_interaction[1] in cell_type_enriched_ids_1:
            #print(temp_interaction[1],"found in",cell_type_1,cell_type_enriched_ids_1)
            gene_source_2+=cell_type_1
        if key_in_dict(temp_interaction[1], cell_type_enriched_dict_2):
        #if temp_interaction[1] in cell_type_enriched_ids_2:
            #print(temp_interaction[1],"found in",cell_type_2,cell_type_enriched_ids_2)
            if gene_source_2!='':
                gene_source_2+='&'
            gene_source_2+=cell_type_2
        ##print(gene_source_2)
        ## double check that there was a gene source found
        if gene_source_1 == '':
            print("couldn't find gene_source_1")
        if gene_source_2 == '':
            print("couldn't find gene_source_2")
        ## format the outputs for this line
        if paracrine:
            temp_full_interaction_description = gene_source_1+":"+temp_interaction[0]+";"+gene_source_2+":"+temp_interaction[1]
            #print(temp_full_interaction_description)
        else:
            temp_full_interaction_description = cell_type_1+":"+temp_interaction[0]+";"+cell_type_2+":"+temp_interaction[1]
        if paracrine and gene_source_1!=gene_source_2:
            final_output_list.append(temp_full_interaction_description)
        elif not paracrine:
            final_output_list.append(temp_full_interaction_description)
    if test_time:
        print('\t\t',time()-start)
    ## and return the final line
    if final_output_list == []:
        return('None')
    return(','.join(final_output_list))



def annotate_gprofiler_output(output_gprofiler):
    global cell_type_idx_dict
    print('\nannotating the gprofiler results\nThis might take a minute...')
    new_output_gprofiler = [output_gprofiler[0]+[output_gprofiler[0][-1]]]
    new_output_gprofiler[0][-2] = 'filtered_annotated_gene_list'
    first = True
    count=0
    total_length = len(output_gprofiler)
    for line in output_gprofiler:
        count+=1
        if count%500 == 0:
            print('\tline:',count,'\tpercent:',100*count/total_length,'%')
        if first:
            first = False
            pass
        else:
            ## put the cell types and annotations into the line, then append it to the new output
            ## but only add it for either all autocrine interactions, or for paracrine interactions
            ## only put it into the output if there are 
            temp_line = line[:]
            temp_line.append(temp_line[-1])
            temp_line[-2] = filter_and_annoatate_genes(temp_line[-1],temp_line[0],temp_line[1])
            if temp_line[-2] != "None":
                new_output_gprofiler.append(temp_line)
    return(new_output_gprofiler)


output_gprofiler = get_gprofiler_on_interaction_lists(interaction_lists, bg = all_included_bg)
output_gprofiler = annotate_gprofiler_output(output_gprofiler)
write_table(output_gprofiler,args.out_dir+'/all_cell_type_specific_interactions_gprofiler.txt')
##########################################################

def get_subset(out_file,term_a,term_b):
    global interaction_title_line
    temp_file_str = out_file
    cmd('rm '+temp_file_str)
    extra_cell_file = open(temp_file_str,'a')
    temp_title_line='\t'.join(interaction_title_line[0])
    extra_cell_file.write(temp_title_line+'\n')
    #extracellular_extracellular_interactions = interaction_title_line[:]
    for i in range(1,len(out_table)):
        if (out_table[i][4]  == term_a and out_table[i][8] == term_b) or (out_table[i][4]  == term_b and out_table[i][8] == term_a):
            temp_line = out_table[i][:]
            temp_line = '\t'.join(temp_line)
            if i!=len(out_table)-1:
                temp_line+='\n'
            extra_cell_file.write(temp_line)
    extra_cell_file.close()



##########################################################
## extracellular:extracellular interactions

#extracellular_extracellular_lists = get_empty_interaction_lists()
#receptor_gene_lists, secreted_gene_lists

# ## go through the out_table and filter it for extracellular:extracellular interactions
# temp_file_str = args.out_dir+'/extracellular_extracellular_cell_type_specific_interactions.txt'
# cmd('rm '+temp_file_str)
# extra_cell_file = open(temp_file_str,'a')
# extra_cell_file.write('\t'.join(interaction_title_line[:])+'\n')
# #extracellular_extracellular_interactions = interaction_title_line[:]
# for i in range(1,len(out_table)):
#     if out_table[i][4]  == 'extracellular' and out_table[i][8] == 'extracellular':
#         temp_line = out_table[i][:]
#         temp_line = '\t'.join(temp_line)
#         if i!=len(out_table)-1:
#             temp_line+='\n'
#         extra_cell_file.write(temp_line)
#         #extracellular_extracellular_interactions.append(out_table[i])
# # extracellular_extracellular_interactions_copy = extracellular_extracellular_interactions[:]
# # write_table(extracellular_extracellular_interactions_copy,args.out_dir+'/extracellular_extracellular_cell_type_specific_interactions.txt')

print('\n\nwriting output files\n')

print('extracellular','extracellular')
temp_file_str = args.out_dir+'/extracellular_extracellular_cell_type_specific_interactions.txt'
get_subset(temp_file_str,'extracellular','extracellular')
print('plasma_membrane','plasma_membrane')
temp_file_str = args.out_dir+'/plasma_membrane_plasma_membrane_cell_type_specific_interactions.txt'
get_subset(temp_file_str,'plasma_membrane','plasma_membrane')

##########################################################
## extracellular:plasma membrane interactions
print('extracellular','plasma_membrane')
def get_unique_ids(i):
    global out_table
    gene_1 = out_table[i][2]
    cell_type_1 = out_table[i][1]
    gene_2 = out_table[i][6]
    cell_type_1 = out_table[i][5]
    inuqe_id_1 = cell_type_1+"|"+gene_1+":"+cell_type_2+"|"+gene_2
    inuqe_id_2 = cell_type_2+"|"+gene_2+":"+cell_type_1+"|"+gene_1
    return(inuqe_id_1,inuqe_id_2)

prev_interact_dict = {}
extracellular_pm_interactions = interaction_title_line[:]
for i in range(1,len(out_table)):
    inuqe_id_1,inuqe_id_2 = get_unique_ids(i)
    if not key_in_dict(inuqe_id_1,prev_interact_dict) and not (key_in_dict(inuqe_id_2,prev_interact_dict)):# temp_line not in extracellular_pm_interactions:
        if out_table[i][4]  == 'extracellular' and out_table[i][8] == 'plasma_membrane':
            extracellular_pm_interactions.append(out_table[i])
            prev_interact_dict[inuqe_id_1]=True
            prev_interact_dict[inuqe_id_2]=True
        elif out_table[i][4]  == 'plasma_membrane' and out_table[i][8] == 'extracellular':
            ## make it a directional file so that things only flow from extracellular to plasma membrane
            temp_line = out_table[i][:]
            temp_cell_description = temp_line[1:5]
            temp_line[1:5] = temp_line[5:9]
            temp_line[5:9] = temp_cell_description
            #if not key_in_dict(inuqe_id_1,prev_interact_dict) and not (key_in_dict(inuqe_id_2,prev_interact_dict)):# temp_line not in extracellular_pm_interactions:
            extracellular_pm_interactions.append(temp_line)
            prev_interact_dict[inuqe_id_1]=True
            prev_interact_dict[inuqe_id_2]=True


extracellular_pm_interactions_copy = extracellular_pm_interactions[:]
write_table(extracellular_pm_interactions_copy,args.out_dir+'/extracellular_plasma_membrane_cell_type_specific_interactions.txt')





# ##########################################################
# ## plasma membrane:plasma membrane interactions

# pm_pm_interactions = interaction_title_line[:]
# for i in range(1,len(out_table)):
#     if out_table[i][4]  == 'plasma_membrane' and out_table[i][8] == 'plasma_membrane':
#         pm_pm_interactions.append(out_table[i])


# pm_pm_interactions_copy = pm_pm_interactions[:]
# write_table(pm_pm_interactions_copy,args.out_dir+'/plasma_membrane_plasma_membrane_cell_type_specific_interactions.txt')
